#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import scipy.signal

__all__ = ['calc_discount_sum_rewards', 'calc_gae']


def calc_discount_sum_rewards(rewards, gamma):
    """ Calculate discounted forward sum of a sequence at each point. 

    Args:
        rewards (List/Tuple/np.array): rewards of (s_t, s_{t+1}, ..., s_T)
        gamma (Scalar): gamma coefficient

    Returns:
        np.array: discounted sum rewards of (s_t, s_{t+1}, ..., s_T)
    """
    return scipy.signal.lfilter([1.0], [1.0, -gamma], rewards[::-1])[::-1]


def calc_gae(rewards, values, next_value, gamma, lam):
    """ Calculate generalized advantage estimator (GAE).
    See: https://arxiv.org/pdf/1506.02438.pdf

    Args:
        rewards (List/Tuple/np.array): rewards of (s_t, s_{t+1}, ..., s_T)
        values (List/Tuple/np.array): values of (s_t, s_{t+1}, ..., s_T)
        next_value (Scalar): value of s_{T+1}
        gamma (Scalar): gamma coefficient
        lam (Scalar): lambda coefficient

    Returns:
        advantages (np.array): advantages of (s_t, s_{t+1}, ..., s_T)
    """
    # temporal differences
    tds = rewards + gamma * np.append(values[1:], next_value) - values
    advantages = calc_discount_sum_rewards(tds, gamma * lam)
    return advantages



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six
from copy import deepcopy

__all__ = ['AttrHolder']


class AttrHolder(object):
    """ Mainly used for maintaining all ParamAttr in a parl.layers.LayerFunc
    """

    def __init__(self, **kwargs):
        """
        Args:
            kwargs: {name:attr}
        """
        self._attrs_dict = {}
        for k, v in six.iteritems(kwargs):
            self._add_attr(k, v)

    def _add_attr(self, name, attr):
        assert name not in self._attrs_dict
        self._attrs_dict[name] = attr

    def __setattr__(self, name, attr):
        if not name.startswith('_'):
            self._add_attr(name, attr)
        else:
            # self._attrs_dict
            super(AttrHolder, self).__setattr__(name, attr)

    def __getattr__(self, name):
        if name in self._attrs_dict.keys():
            return self._attrs_dict[name]
        else:
            return None

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in six.iteritems(self.__dict__):
            setattr(result, k, deepcopy(v, memo))
        return result

    def sorted(self):
        """
        Returns:
            list of all attrs, which is sorted by key
        """
        return [self._attrs_dict[k] for k in sorted(self._attrs_dict.keys())]

    def tolist(self):
        """
        Returns:
            list of all attrs
        """
        return list(six.itervalues(self._attrs_dict))



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from torch.distributions import Categorical
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.optim import lr_scheduler
from random import random, randint

import parl
from parl.utils.scheduler import PiecewiseScheduler, LinearDecayScheduler

__all__ = ['A2C']


class A2C(parl.Algorithm):
    def __init__(self, model, config):
        assert isinstance(config['vf_loss_coeff'], (int, float))
        self.model = model
        self.vf_loss_coeff = config['vf_loss_coeff']
        self.optimizer = optim.Adam(
            self.model.parameters(), lr=config['learning_rate'])
        self.config = config

        self.lr_scheduler = LinearDecayScheduler(config['start_lr'],
                                                 config['max_sample_steps'])

        self.entropy_coeff_scheduler = PiecewiseScheduler(
            config['entropy_coeff_scheduler'])

    def learn(self, obs, actions, advantages, target_values):
        prob = self.model.policy(obs, softmax_dim=1)
        policy_distri = Categorical(prob)
        actions_log_probs = policy_distri.log_prob(actions)

        # The policy gradient loss
        pi_loss = -((actions_log_probs * advantages).sum())

        # The value function loss
        values = self.model.value(obs).reshape(-1)
        delta = values - target_values
        vf_loss = 0.5 * torch.mul(delta, delta).sum()

        # The entropy loss (We want to maximize entropy, so entropy_ceoff < 0)
        policy_entropy = policy_distri.entropy()
        entropy = policy_entropy.sum()

        lr = self.lr_scheduler.step(step_num=obs.shape[0])
        entropy_coeff = self.entropy_coeff_scheduler.step()

        total_loss = pi_loss + vf_loss * self.vf_loss_coeff + entropy * entropy_coeff

        for param_group in self.optimizer.param_groups:
            param_group['lr'] = lr

        total_loss.backward()
        self.optimizer.step()
        self.optimizer.zero_grad()

        return total_loss, pi_loss, vf_loss, entropy, lr, entropy_coeff

    def sample(self, obs):
        prob, values = self.model.policy_and_value(obs)
        sample_actions = Categorical(prob).sample()

        return sample_actions, values

    def predict(self, obs):
        prob = self.model.policy(obs)
        _, predict_actions = prob.max(-1)

        return predict_actions

    def value(self, obs):
        values = self.model.value(obs)
        return values



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
warnings.simplefilter('default')

import paddle.fluid as fluid
from parl.core.fluid import layers
from parl.core.fluid.algorithm import Algorithm
from parl.core.fluid.policy_distribution import CategoricalDistribution

__all__ = ['A3C']


class A3C(Algorithm):
    def __init__(self, model, vf_loss_coeff=None):
        """ A3C/A2C algorithm
        
        Args:
            model (parl.Model): forward network of policy and value
            vf_loss_coeff (float): coefficient of the value function loss
        """

        self.model = model
        assert isinstance(vf_loss_coeff, (int, float))
        self.vf_loss_coeff = vf_loss_coeff

    def learn(self, obs, actions, advantages, target_values, learning_rate,
              entropy_coeff):
        """
        Args:
            obs: An float32 tensor of shape ([B] + observation_space).
                 E.g. [B, C, H, W] in atari.
            actions: An int64 tensor of shape [B].
            advantages: A float32 tensor of shape [B].
            target_values: A float32 tensor of shape [B].
            learning_rate: float scalar of learning rate.
            entropy_coeff: float scalar of entropy coefficient.
        """
        logits = self.model.policy(obs)
        policy_distribution = CategoricalDistribution(logits)
        actions_log_probs = policy_distribution.logp(actions)

        # The policy gradient loss
        pi_loss = -1.0 * layers.reduce_sum(actions_log_probs * advantages)

        # The value function loss
        values = self.model.value(obs)
        delta = values - target_values
        vf_loss = 0.5 * layers.reduce_sum(layers.square(delta))

        # The entropy loss (We want to maximize entropy, so entropy_ceoff < 0)
        policy_entropy = policy_distribution.entropy()
        entropy = layers.reduce_sum(policy_entropy)

        total_loss = (
            pi_loss + vf_loss * self.vf_loss_coeff + entropy * entropy_coeff)

        reduced_total_loss = layers.reduce_sum(total_loss)

        fluid.clip.set_gradient_clip(
            clip=fluid.clip.GradientClipByGlobalNorm(clip_norm=40.0))

        optimizer = fluid.optimizer.AdamOptimizer(learning_rate)
        optimizer.minimize(reduced_total_loss)

        return reduced_total_loss, pi_loss, vf_loss, entropy

    def sample(self, obs):
        """
        Args:
            obs: An float32 tensor of shape ([B] + observation_space).
                 E.g. [B, C, H, W] in atari.
        """
        logits, values = self.model.policy_and_value(obs)

        policy_dist = CategoricalDistribution(logits)
        sample_actions = policy_dist.sample()

        return sample_actions, values

    def predict(self, obs):
        """
        Args:
            obs: An float32 tensor of shape ([B] + observation_space).
                 E.g. [B, C, H, W] in atari.
        """
        logits = self.model.policy(obs)
        probs = layers.softmax(logits)

        predict_actions = layers.argmax(probs, 1)

        return predict_actions

    def value(self, obs):
        """
        Args:
            obs: An float32 tensor of shape ([B] + observation_space).
                 E.g. [B, C, H, W] in atari.
        """
        values = self.model.value(obs)
        return values



#   Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import parl
import copy
from copy import deepcopy
from paddle.fluid import layers
import paddle.fluid as fluid

__all__ = ['QMIX']


class QMIX(parl.Algorithm):
    def __init__(self, agent_model, qmixer_model, config):
        """ QMIX algorithm
        Args:
            agent_model (parl.Model): agents' local q network for decision making.
            qmixer_model (parl.Model): A mixing network which takes local q values as input
            config (dict): config of the algorithm.
        """
        self.agent_model = agent_model
        self.qmixer_model = qmixer_model
        self.target_agent_model = deepcopy(self.agent_model)
        self.target_qmixer_model = deepcopy(self.qmixer_model)

        self.lr = config['lr']
        self.gamma = config['gamma']
        self.double_q = config['double_q']
        self.n_agents = config['n_agents']
        self.n_actions = config['n_actions']
        self.obs_shape = config['obs_shape']
        self.batch_size = config['batch_size']
        self.episode_limit = config['episode_limit']
        self.rnn_hidden_dim = config['rnn_hidden_dim']
        self.clip_grad_norm = config['clip_grad_norm']
        assert isinstance(self.gamma, float)
        assert isinstance(self.lr, float)

    def predict_local_q(self, obs, hidden_state):
        ''' Predict local q values for each agent.
        Args:
            obs:           (n_agents, obs_shape)
            hidden_states: (n_agents, rnn_hidden_dim)
        Returns:
            self.agent_model(obs, hidden_state)
        '''
        return self.agent_model(obs, hidden_state)

    def learn(self, init_hidden_states, target_init_hidden_states, state_batch,
              actions_batch, reward_batch, terminated_batch, obs_batch,
              available_actions_batch, filled_batch):
        '''
        Args:
            init_hidden_states:       (batch_size, n_agents, rnn_hidden_dim)
            target_init_hidden_states:(batch_size, n_agents, rnn_hidden_dim)
            state_batch:              (batch_size, T, state_shape)
            actions_batch:            (batch_size, T, n_agents)
            reward_batch:             (batch_size, T, 1)
            terminated_batch:         (batch_size, T, 1)
            obs_batch:                (batch_size, T, n_agents, obs_shape)
            available_actions_batch:  (batch_size, T, n_agents, n_actions)
            filled_batch:             (batch_size, T, 1)
        Returns:
            loss (float): train loss
            td_erro (float): train TD error
        '''
        hidden_states = init_hidden_states
        target_hidden_states = target_init_hidden_states

        reward_batch = layers.slice(
            reward_batch, axes=[1], starts=[0], ends=[-1])
        actions_batch = layers.unsqueeze(
            layers.slice(actions_batch, axes=[1], starts=[0], ends=[-1]),
            axes=[-1])
        terminated_batch = layers.slice(
            terminated_batch, axes=[1], starts=[0], ends=[-1])
        filled_batch = layers.slice(
            filled_batch, axes=[1], starts=[0], ends=[-1])

        mask = (1 - filled_batch) * (1 - terminated_batch)

        local_qs = []
        target_local_qs = []
        for t in range(self.episode_limit):
            obs = obs_batch[:, t, :, :]
            obs = layers.reshape(
                obs, shape=(self.batch_size * self.n_agents, self.obs_shape))
            hidden_states = layers.reshape(
                hidden_states,
                shape=(self.batch_size * self.n_agents, self.rnn_hidden_dim))
            local_q, hidden_states = self.agent_model(obs, hidden_states)
            local_q = layers.reshape(
                local_q,
                shape=(self.batch_size, self.n_agents, self.n_actions))
            local_qs.append(local_q)

            target_hidden_states = layers.reshape(
                target_hidden_states,
                shape=(self.batch_size * self.n_agents, self.rnn_hidden_dim))
            target_local_q, target_hidden_states = self.target_agent_model(
                obs, target_hidden_states)
            target_local_q = layers.reshape(
                target_local_q,
                shape=(self.batch_size, self.n_agents, self.n_actions))
            target_local_qs.append(target_local_q)

        local_qs = layers.stack(local_qs, axis=1)
        target_local_qs = layers.stack(target_local_qs[1:], axis=1)

        actions_batch_one_hot = layers.one_hot(
            actions_batch, depth=self.n_actions)
        actions_batch_one_hot = layers.cast(
            actions_batch_one_hot, dtype='float32')
        sliced_local_qs = layers.slice(
            local_qs, axes=[1], starts=[0], ends=[-1])
        chosen_action_local_qs = layers.reduce_sum(
            layers.elementwise_mul(actions_batch_one_hot, sliced_local_qs),
            dim=-1)

        available_actions_batch = layers.cast(
            available_actions_batch, dtype='float32')
        action_mask = -1e10 * (1.0 - available_actions_batch)
        target_action_mask = layers.slice(
            action_mask, axes=[1], starts=[1], ends=[10000])
        target_action_zero_mask = layers.slice(
            available_actions_batch, axes=[1], starts=[1], ends=[10000])
        # mask unavailable actions
        target_local_qs = target_local_qs * target_action_zero_mask + target_action_mask

        if self.double_q:
            # mask unavailable actions
            masked_local_qs = local_qs * available_actions_batch + action_mask
            masked_local_qs.stop_gradient = True
            masked_local_qs = layers.slice(
                masked_local_qs, axes=[1], starts=[1], ends=[100000])
            cur_max_actions = layers.unsqueeze(
                layers.argmax(masked_local_qs, axis=-1), axes=-1)
            cur_max_actions_one_hot = layers.one_hot(
                cur_max_actions, depth=self.n_actions)
            cur_max_actions_one_hot = layers.cast(
                cur_max_actions_one_hot, dtype='float32')
        else:
            cur_max_actions = layers.unsqueeze(
                layers.argmax(target_local_qs, axis=-1), axes=-1)
            cur_max_actions_one_hot = layers.one_hot(
                cur_max_actions, depth=self.n_actions)
            cur_max_actions_one_hot = layers.cast(
                cur_max_actions_one_hot, dtype='float32')

        target_local_max_qs = layers.reduce_sum(
            layers.elementwise_mul(cur_max_actions_one_hot, target_local_qs),
            dim=-1)

        chosen_action_global_qs = self.qmixer_model(
            chosen_action_local_qs,
            layers.slice(state_batch, axes=[1], starts=[0], ends=[-1]))
        target_global_max_qs = self.target_qmixer_model(
            target_local_max_qs,
            layers.slice(state_batch, axes=[1], starts=[1], ends=[10000]))

        target = reward_batch + self.gamma * (
            1 - terminated_batch) * target_global_max_qs
        target.stop_gradient = True

        td_error = target - chosen_action_global_qs
        masked_td_error = td_error * mask
        mean_td_error = layers.reduce_sum(masked_td_error) / layers.reduce_sum(
            mask)

        loss = layers.reduce_sum(
            layers.square(masked_td_error)) / layers.reduce_sum(mask)
        if self.clip_grad_norm:
            clip = fluid.clip.GradientClipByGlobalNorm(
                clip_norm=self.clip_grad_norm)
            optimizer = fluid.optimizer.RMSPropOptimizer(
                learning_rate=self.lr, rho=0.99, epsilon=1e-6, grad_clip=clip)
        else:
            optimizer = fluid.optimizer.RMSPropOptimizer(
                learning_rate=self.lr, rho=0.99, epsilon=1e-6)

        optimizer.minimize(loss)
        return loss, mean_td_error

    def sync_target(self):
        '''sync weights of model to target model
        '''
        self.agent_model.sync_weights_to(self.target_agent_model)
        self.qmixer_model.sync_weights_to(self.target_qmixer_model)



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from parl.utils.window_stat import WindowStat

__all_ = ['TimeStat']


class TimeStat(object):
    """A time stat for logging the elapsed time of code running

    Example:
        time_stat = TimeStat()
        with time_stat:
            // some code
        print(time_stat.mean)
    """

    def __init__(self, window_size=1):
        self.time_samples = WindowStat(window_size)
        self._start_time = None

    def __enter__(self):
        self._start_time = time.time()

    def __exit__(self, type, value, tb):
        time_delta = time.time() - self._start_time
        self.time_samples.add(time_delta)

    @property
    def mean(self):
        return self.time_samples.mean

    @property
    def min(self):
        return self.time_samples.min

    @property
    def max(self):
        return self.time_samples.max



#   Copyright (c) 2021 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import parl
import paddle
import paddle.nn.functional as F
from copy import deepcopy

__all__ = ['TD3']


class TD3(parl.Algorithm):
    def __init__(self,
                 model,
                 gamma=None,
                 tau=None,
                 actor_lr=None,
                 critic_lr=None,
                 policy_noise=0.2,
                 noise_clip=0.5,
                 policy_freq=2):
        """ TD3 algorithm

        Args:
            model(parl.Model): forward network of actor and critic.
            gamma(float): discounted factor for reward computation
            tau (float): decay coefficient when updating the weights of self.target_model with self.model
            actor_lr (float): learning rate of the actor model
            critic_lr (float): learning rate of the critic model
            policy_noise(float): noise added to target policy during critic update
            noise_clip(float): range to clip target policy noise
            policy_freq(int): frequency of delayed policy updates
        """
        assert isinstance(gamma, float)
        assert isinstance(tau, float)
        assert isinstance(actor_lr, float)
        assert isinstance(critic_lr, float)

        self.gamma = gamma
        self.tau = tau
        self.actor_lr = actor_lr
        self.critic_lr = critic_lr
        self.policy_noise = policy_noise
        self.noise_clip = noise_clip
        self.policy_freq = policy_freq
        self.total_it = 0

        self.model = model
        self.target_model = deepcopy(self.model)
        self.actor_optimizer = paddle.optimizer.Adam(
            learning_rate=actor_lr, parameters=self.model.get_actor_params())
        self.critic_optimizer = paddle.optimizer.Adam(
            learning_rate=critic_lr, parameters=self.model.get_critic_params())

    def predict(self, obs):
        action = self.model.policy(obs)
        return action

    def learn(self, obs, action, reward, next_obs, terminal):
        self.total_it += 1
        critic_loss = self._critic_learn(obs, action, reward, next_obs,
                                         terminal)
        if self.total_it % self.policy_freq == 0:
            actor_loss = self._actor_learn(obs)
        return critic_loss

    def _critic_learn(self, obs, action, reward, next_obs, terminal):
        with paddle.no_grad():
            noise = paddle.randn(action.shape) * self.policy_noise
            noise = paddle.clip(noise, -self.noise_clip, self.noise_clip)

            next_action = self.target_model.policy(next_obs) + noise
            next_action = paddle.clip(next_action, -1., 1.)

            target_q1, target_q2 = self.target_model.value(
                next_obs, next_action)
            target_q = paddle.minimum(target_q1, target_q2)
            terminal = paddle.cast(terminal, dtype='float32')
            target_q = reward + (1. - terminal) * self.gamma * target_q
        current_q1, current_q2 = self.model.value(obs, action)

        critic_loss = F.mse_loss(current_q1, target_q) + F.mse_loss(
            current_q2, target_q)

        self.critic_optimizer.clear_grad()
        critic_loss.backward()
        self.critic_optimizer.step()
        return critic_loss

    def _actor_learn(self, obs):
        actor_loss = -self.model.Q1(obs, self.model.policy(obs)).mean()

        self.actor_optimizer.clear_grad()
        actor_loss.backward()
        self.actor_optimizer.step()

        self.sync_target()
        return actor_loss

    def sync_target(self, decay=None):
        if decay is None:
            decay = 1. - self.tau
        self.model.sync_weights_to(self.target_model, decay=decay)



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from parl.core.fluid import layers
from copy import deepcopy
import numpy as np
from paddle import fluid
from paddle.fluid.layers import Normal
from parl.core.fluid.algorithm import Algorithm

epsilon = 1e-6

__all__ = ['SAC']


class SAC(Algorithm):
    def __init__(self,
                 actor,
                 critic,
                 max_action,
                 alpha=0.2,
                 gamma=None,
                 tau=None,
                 actor_lr=None,
                 critic_lr=None):
        """ SAC algorithm

        Args:
            actor (parl.Model): forward network of actor.
            critic (patl.Model): forward network of the critic.
            max_action (float): the largest value that an action can be, env.action_space.high[0]
            alpha (float): Temperature parameter determines the relative importance of the entropy against the reward
            gamma (float): discounted factor for reward computation.
            tau (float): decay coefficient when updating the weights of self.target_model with self.model
            actor_lr (float): learning rate of the actor model
            critic_lr (float): learning rate of the critic model
        """
        assert isinstance(gamma, float)
        assert isinstance(tau, float)
        assert isinstance(actor_lr, float)
        assert isinstance(critic_lr, float)
        assert isinstance(alpha, float)
        self.max_action = max_action
        self.gamma = gamma
        self.tau = tau
        self.actor_lr = actor_lr
        self.critic_lr = critic_lr

        self.alpha = alpha

        self.actor = actor
        self.critic = critic
        self.target_critic = deepcopy(critic)

    def predict(self, obs):
        """ use actor model of self.policy to predict the action
        """
        mean, _ = self.actor.policy(obs)
        mean = layers.tanh(mean) * self.max_action
        return mean

    def sample(self, obs):
        mean, log_std = self.actor.policy(obs)
        std = layers.exp(log_std)
        normal = Normal(mean, std)
        x_t = normal.sample([1])[0]
        y_t = layers.tanh(x_t)
        action = y_t * self.max_action
        log_prob = normal.log_prob(x_t)
        log_prob -= layers.log(self.max_action * (1 - layers.pow(y_t, 2)) +
                               epsilon)
        log_prob = layers.reduce_sum(log_prob, dim=1, keep_dim=True)
        log_prob = layers.squeeze(log_prob, axes=[1])
        return action, log_prob

    def learn(self, obs, action, reward, next_obs, terminal):
        actor_cost = self.actor_learn(obs)
        critic_cost = self.critic_learn(obs, action, reward, next_obs,
                                        terminal)
        return critic_cost, actor_cost

    def actor_learn(self, obs):
        action, log_pi = self.sample(obs)
        qf1_pi, qf2_pi = self.critic.value(obs, action)
        min_qf_pi = layers.elementwise_min(qf1_pi, qf2_pi)
        cost = log_pi * self.alpha - min_qf_pi
        cost = layers.reduce_mean(cost)
        optimizer = fluid.optimizer.AdamOptimizer(self.actor_lr)
        optimizer.minimize(cost, parameter_list=self.actor.parameters())

        return cost

    def critic_learn(self, obs, action, reward, next_obs, terminal):
        next_obs_action, next_obs_log_pi = self.sample(next_obs)
        qf1_next_target, qf2_next_target = self.target_critic.value(
            next_obs, next_obs_action)
        min_qf_next_target = layers.elementwise_min(
            qf1_next_target, qf2_next_target) - next_obs_log_pi * self.alpha

        terminal = layers.cast(terminal, dtype='float32')
        target_Q = reward + (1.0 - terminal) * self.gamma * min_qf_next_target
        target_Q.stop_gradient = True

        current_Q1, current_Q2 = self.critic.value(obs, action)
        cost = layers.square_error_cost(current_Q1,
                                        target_Q) + layers.square_error_cost(
                                            current_Q2, target_Q)
        cost = layers.reduce_mean(cost)
        optimizer = fluid.optimizer.AdamOptimizer(self.critic_lr)
        optimizer.minimize(cost)
        return cost

    def sync_target(self, decay=None):
        if decay is None:
            decay = 1.0 - self.tau
        self.critic.sync_weights_to(self.target_critic, decay=decay)



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from tensorboardX import SummaryWriter
from parl.utils import logger
from parl.utils.machine_info import get_ip_address

__all__ = []

_writer = None
_WRITTER_METHOD = ['add_scalar', 'add_histogram', 'close', 'flush']


def create_file_after_first_call(func_name):
    def call(*args, **kwargs):
        global _writer
        if _writer is None:
            logdir = logger.get_dir()
            if logdir is None:
                logdir = logger.auto_set_dir(action='d')
                logger.warning(
                    "[tensorboard] logdir is None, will save tensorboard files to {}\nView the data using: tensorboard --logdir=./{} --host={}"
                    .format(logdir, logdir, get_ip_address()))
            _writer = SummaryWriter(logdir=logger.get_dir())
        func = getattr(_writer, func_name)
        func(*args, **kwargs)
        _writer.flush()

    return call


# export writter functions
for func_name in _WRITTER_METHOD:
    locals()[func_name] = create_file_after_first_call(func_name)
    __all__.append(func_name)



#   Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import grpc
import time
import threading
from parl.remote import remote_constants
from parl.remote.grpc_heartbeat import heartbeat_pb2
from parl.remote.grpc_heartbeat import heartbeat_pb2_grpc
from parl.utils import logger


class HeartbeatClientThread(threading.Thread):
    def __init__(self,
                 heartbeat_server_addr,
                 heartbeat_exit_callback_func,
                 exit_func_args=(),
                 exit_func_kwargs={}):
        """Create a thread to run the heartbeat client.

            Args:
                heartbeat_server_addr(str): the address of the heartbeat server.
                heartbeat_exit_callback_func(function): A callback function, which will be called after the 
                                                        heartbeat exit.
                exit_func_args(tuple): the argument tuple for calling the heartbeat_exit_callback_func. Defaults to ().
                exit_func_kwargs(dict): the argument dict for calling the heartbeat_exit_callback_func. Defaults to {}.
        """
        assert isinstance(heartbeat_server_addr, str)
        assert callable(
            heartbeat_exit_callback_func), "It should be a function."
        assert isinstance(exit_func_args, tuple)
        assert isinstance(exit_func_kwargs, dict)

        threading.Thread.__init__(self)
        self.heartbeat_server_addr = heartbeat_server_addr

        self.heartbeat_exit_callback_func = heartbeat_exit_callback_func
        self._exit_func_args = exit_func_args
        self._exit_func_kwargs = exit_func_kwargs

        self.exit_flag = False

    def exit(self):
        self.exit_flag = True

    def run(self):
        # unset http_proxy and https_proxy
        if 'http_proxy' in os.environ:
            del os.environ['http_proxy']
        if 'https_proxy' in os.environ:
            del os.environ['https_proxy']

        with grpc.insecure_channel(
                self.heartbeat_server_addr,
                options=[('grpc.max_receive_message_length', -1),
                         ('grpc.max_send_message_length', -1)]) as channel:
            stub = heartbeat_pb2_grpc.GrpcHeartbeatStub(channel)

            while True:
                if self.exit_flag:
                    break

                try:
                    response = stub.Send(
                        heartbeat_pb2.Request(
                            tag=remote_constants.HEARTBEAT_TAG),
                        timeout=remote_constants.HEARTBEAT_RCVTIMEO_S)

                    if response.tag == remote_constants.HEARTBEAT_TAG:
                        pass
                    elif response.tag == remote_constants.HEARTBEAT_OUT_OF_MEMORY_TAG:
                        logger.error(response.extra_message)
                        break
                    else:
                        raise NotImplementedError

                except grpc._channel._InactiveRpcError as e:
                    break

                time.sleep(remote_constants.HEARTBEAT_INTERVAL_S)

        # heartbeat is exit, call the exit function.
        self.heartbeat_exit_callback_func(*self._exit_func_args,
                                          **self._exit_func_kwargs)



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os
import subprocess
import numpy as np
from parl.utils import logger

__all__ = [
    'has_func', 'to_str', 'to_byte', '_IS_PY2', '_IS_PY3', 'MAX_INT32',
    '_HAS_FLUID', '_HAS_PADDLE', '_HAS_TORCH', '_IS_WINDOWS', '_IS_MAC',
    'kill_process', 'get_fluid_version', 'isnotebook', 'check_version_for_xpu',
    'check_version_for_fluid'
]


def has_func(obj, fun):
    """check if a class has specified function: https://stackoverflow.com/a/5268474

    Args:
        obj: the class to check
        fun: specified function to check
    Returns:
        A bool to indicate if obj has funtion "fun"
    """
    check_fun = getattr(obj, fun, None)
    return callable(check_fun)


def to_str(byte):
    """ convert byte to string in pytohn2/3
    """
    return str(byte.decode())


def to_byte(string):
    """ convert byte to string in pytohn2/3
    """
    return string.encode()


_IS_PY2 = (sys.version_info[0] == 2)
_IS_PY3 = (sys.version_info[0] == 3)


def get_fluid_version():
    import paddle
    paddle_version = int(paddle.__version__.replace('.', '').split('-')[0])
    return paddle_version


MAX_INT32 = 0x7fffffff

try:
    _HAS_FLUID = False
    _HAS_PADDLE = False
    import paddle
    from paddle import fluid

    paddle_version = get_fluid_version()
    logger.info("paddlepaddle version: {}.".format(paddle.__version__))
    if paddle_version < 200 and paddle_version != 0:
        assert paddle_version >= 185, "PARL requires paddle >= 1.8.5 and paddle < 2.0.0"
        _HAS_FLUID = True
    else:
        _HAS_PADDLE = True
except ImportError as e:
    _HAS_FLUID = False
    _HAS_PADDLE = False
    if _IS_PY2 and "{}".format(e) != "No module named paddle":
        logger.warning("import paddle error:\n{}".format(e))
    if _IS_PY3 and "{}".format(e) != "No module named 'paddle'":
        logger.warning("import paddle error:\n{}".format(e))

try:
    import torch
    _HAS_TORCH = True
except ImportError:
    _HAS_TORCH = False

_IS_WINDOWS = (sys.platform == 'win32')
_IS_MAC = (sys.platform == 'darwin')


def kill_process(regex_pattern):
    """kill process whose execution commnad is matched by regex pattern

    Args:
        regex_pattern(string): regex pattern used to filter the process to be killed
    
    NOTE:
        In windows, we will replace sep `/` with `\\\\`
    """
    if _IS_WINDOWS:
        regex_pattern = regex_pattern.replace('/', '\\\\')
        command = r'''for /F "skip=2 tokens=2 delims=," %a in ('wmic process where "commandline like '%{}%'" get processid^,status /format:csv') do taskkill /F /T /pid %a'''.format(
            regex_pattern)
        with os.popen(command) as p:
            p.read()
    else:
        command = "ps aux | grep {} | awk '{{print $2}}' | xargs kill -9".format(
            regex_pattern)
        subprocess.call([command], shell=True)


def isnotebook():
    """check if the code is excuted in the IPython notebook
    Reference: https://stackoverflow.com/a/39662359
    """
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True  # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


def check_version_for_xpu():
    """check paddle version if the code requires to run on xpu
    """
    err = "To use xpu, PARL requires paddle version >= 2.0.0. " \
          "Please make sure the version is consistent with your code."
    import paddle
    from paddle import fluid

    paddle_version = get_fluid_version()
    assert paddle_version >= 200 or paddle_version == 0, err


def check_version_for_fluid():
    """check the paddle version if the code requires fluid-based implementation.
    """
    err = "To use fluid version of examples, PARL requires paddle version < 2.0.0. " \
          "Please make sure the version is consistent with your code."
    import paddle
    from paddle import fluid

    paddle_version = get_fluid_version()
    assert paddle_version >= 185 and paddle_version < 200, err



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
warnings.simplefilter('default')

import copy
import numpy as np
import paddle.fluid as fluid
from parl.core.fluid.algorithm import Algorithm
from parl.core.fluid import layers

__all__ = ['DDQN']


class DDQN(Algorithm):
    def __init__(self, model, act_dim=None, gamma=None, lr=None):
        """ Double DQN algorithm
        Args:
            model (parl.Model): model defining forward network of Q function
            act_dim (int): dimension of the action space
            gamma (float): discounted factor for reward computation.
            lr (float): learning rate.
        """
        self.model = model
        self.target_model = copy.deepcopy(model)

        assert isinstance(act_dim, int)
        assert isinstance(gamma, float)

        self.act_dim = act_dim
        self.gamma = gamma
        self.lr = lr

    def predict(self, obs):
        """ use value model self.model to predict the action value
        """
        return self.model.value(obs)

    def learn(self,
              obs,
              action,
              reward,
              next_obs,
              terminal,
              learning_rate=None):
        """ update value model self.model with DQN algorithm
        """
        # Support the modification of learning_rate
        if learning_rate is None:
            assert isinstance(
                self.lr,
                float), "Please set the learning rate of DQN in initializaion."
            learning_rate = self.lr

        pred_value = self.model.value(obs)
        action_onehot = layers.one_hot(action, self.act_dim)
        action_onehot = layers.cast(action_onehot, dtype='float32')
        pred_action_value = layers.reduce_sum(
            layers.elementwise_mul(action_onehot, pred_value), dim=1)

        # calculate the target q value
        next_action_value = self.model.value(next_obs)
        greedy_action = layers.argmax(next_action_value, axis=-1)
        greedy_action = layers.unsqueeze(greedy_action, axes=[1])
        greedy_action_onehot = layers.one_hot(greedy_action, self.act_dim)
        next_pred_value = self.target_model.value(next_obs)
        max_v = layers.reduce_sum(
            greedy_action_onehot * next_pred_value, dim=1)
        max_v.stop_gradient = True

        target = reward + (
            1.0 - layers.cast(terminal, dtype='float32')) * self.gamma * max_v
        cost = layers.square_error_cost(pred_action_value, target)
        cost = layers.reduce_mean(cost)
        optimizer = fluid.optimizer.Adam(
            learning_rate=learning_rate, epsilon=1e-3)
        optimizer.minimize(cost)
        return cost

    def sync_target(self):
        """ sync weights of self.model to self.target_model
        """
        self.model.sync_weights_to(self.target_model)



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Common functions of PARL framework
"""

import paddle.fluid as fluid
from paddle.fluid.executor import _fetch_var
from parl.utils import machine_info

__all__ = ['fetch_framework_var', 'fetch_value', 'set_value', 'inverse']


def fetch_framework_var(attr_name):
    """ Fetch framework variable according given attr_name.
        Return a new reusing variable through create_parameter way 

    Args:
        attr_name: string, attr name of parameter

    Returns:
        framework_var: framework.Varialbe
    """

    scope = fluid.executor.global_scope()
    core_var = scope.find_var(attr_name)
    if core_var == None:
        raise KeyError(
            "Unable to find the variable:{}. Synchronize parameters before initialization or attr_name does not exist."
            .format(attr_name))
    shape = core_var.get_tensor().shape()
    framework_var = fluid.layers.create_parameter(
        shape=shape, dtype='float32', attr=fluid.ParamAttr(name=attr_name))
    return framework_var


def fetch_value(attr_name):
    """ Given name of ParamAttr, fetch numpy value of the parameter in global_scope
    
    Args:
        attr_name: ParamAttr name of parameter

    Returns:
        numpy.ndarray
    """
    return _fetch_var(attr_name, return_numpy=True)


def set_value(attr_name, value, is_gpu_available):
    """ Given name of ParamAttr, set numpy value to the parameter in global_scope
    
    Args:
        attr_name(string): ParamAttr name of parameter
        value(np.array): numpy value
        is_gpu_available(bool): whether is gpu available
    """
    place = fluid.CUDAPlace(0) if is_gpu_available else fluid.CPUPlace()
    var = _fetch_var(attr_name, return_numpy=False)
    var.set(value, place)


def inverse(x):
    """ Inverse 0/1 variable

    Args:
        x: variable with float32 dtype
    
    Returns:
        inverse_x: variable with float32 dtype
    """
    inverse_x = -1.0 * x + 1.0
    return inverse_x



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from parl.core.fluid import layers
from copy import deepcopy
from paddle import fluid
from parl.core.fluid.algorithm import Algorithm

__all__ = ['TD3']


class TD3(Algorithm):
    def __init__(
            self,
            model,
            max_action,
            gamma=None,
            tau=None,
            actor_lr=None,
            critic_lr=None,
            policy_noise=0.2,  # Noise added to target policy during critic update
            noise_clip=0.5,  # Range to clip target policy noise
            policy_freq=2):  # Frequency of delayed policy updates
        assert isinstance(gamma, float)
        assert isinstance(tau, float)
        assert isinstance(actor_lr, float)
        assert isinstance(critic_lr, float)
        self.max_action = max_action
        self.gamma = gamma
        self.tau = tau
        self.actor_lr = actor_lr
        self.critic_lr = critic_lr
        self.policy_noise = policy_noise
        self.noise_clip = noise_clip
        self.policy_freq = policy_freq

        self.model = model
        self.target_model = deepcopy(model)

    def predict(self, obs):
        """ use actor model of self.model to predict the action
        """
        return self.model.policy(obs)

    def actor_learn(self, obs):
        action = self.model.policy(obs)
        Q = self.model.Q1(obs, action)
        cost = layers.reduce_mean(-1.0 * Q)
        optimizer = fluid.optimizer.AdamOptimizer(self.actor_lr)
        optimizer.minimize(cost, parameter_list=self.model.get_actor_params())
        return cost

    def critic_learn(self, obs, action, reward, next_obs, terminal):
        noise = layers.gaussian_random_batch_size_like(
            action, shape=[-1, action.shape[1]])
        noise = layers.clip(
            noise * self.policy_noise,
            min=-self.noise_clip,
            max=self.noise_clip)
        next_action = self.target_model.policy(next_obs) + noise
        next_action = layers.clip(next_action, -self.max_action,
                                  self.max_action)

        next_Q1, next_Q2 = self.target_model.value(next_obs, next_action)
        next_Q = layers.elementwise_min(next_Q1, next_Q2)

        terminal = layers.cast(terminal, dtype='float32')
        target_Q = reward + (1.0 - terminal) * self.gamma * next_Q
        target_Q.stop_gradient = True

        current_Q1, current_Q2 = self.model.value(obs, action)
        cost = layers.square_error_cost(current_Q1,
                                        target_Q) + layers.square_error_cost(
                                            current_Q2, target_Q)
        cost = layers.reduce_mean(cost)
        optimizer = fluid.optimizer.AdamOptimizer(self.critic_lr)
        optimizer.minimize(cost)
        return cost

    def sync_target(self, decay=None):
        if decay is None:
            decay = 1.0 - self.tau
        self.model.sync_weights_to(self.target_model, decay=decay)



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class ResourceError(Exception):
    """
    No available cpu resources error.
    """

    def __init__(self, error_info):
        self.error_info = error_info

    def __str__(self):
        return self.error_info


class RemoteError(Exception):
    """
    Super class of exceptions in remote module.
    """

    def __init__(self, func_name, error_info):
        self.error_info = "[PARL remote error when calling " +\
            "function `{}`]:\n{}".format(func_name, error_info)

    def __str__(self):
        return self.error_info


class RemoteSerializeError(RemoteError):
    """
    Serialize error from remote
    """

    def __init__(self, func_name, error_info):
        super(RemoteSerializeError, self).__init__(func_name, error_info)

    def __str__(self):
        return self.error_info


class RemoteDeserializeError(RemoteError):
    """
    Deserialize error from remote
    """

    def __init__(self, func_name, error_info):
        super(RemoteDeserializeError, self).__init__(func_name, error_info)

    def __str__(self):
        return self.error_info


class RemoteAttributeError(RemoteError):
    """
    Attribute error from remote
    """

    def __init__(self, func_name, error_info):
        super(RemoteAttributeError, self).__init__(func_name, error_info)

    def __str__(self):
        return self.error_info


class FutureGetRepeatedlyError(Exception):
    """
    Calling the get function of `FutureObject` repeatedly.
    """

    def __init__(self):
        self.error_info = "The `get` function of the `FutureObject` has been called before. We can only call the `get` function for each FutureObject **once**."

    def __str__(self):
        return self.error_info


class FutureFunctionError(Exception):
    """
    Error raised when calling future function.
    """

    def __init__(self, func_name):
        self.error_info = "There is an error raised when calling the future function `{}`.\n".format(func_name) + \
                "You can see the detailed error message above, which is printed by another thread."

    def __str__(self):
        return self.error_info


class FutureObjectEmpty(Exception):
    """
    Exception raised when non-blocking get() (or get_nowait()) is called on a `FutureObject` which is empty.
    """

    def __init__(self):
        self.error_info = "The result is currently unavailable."

    def __str__(self):
        return self.error_info



#   Copyright (c) 2021 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import paddle
import paddle.nn as nn
import parl
from copy import deepcopy
import paddle.nn.functional as F

__all__ = ['QMIX']


class QMIX(parl.Algorithm):
    def __init__(self,
                 agent_model,
                 qmixer_model,
                 double_q=True,
                 gamma=0.99,
                 lr=0.0005,
                 clip_grad_norm=None):
        """ QMIX algorithm

        Args:
            agent_model (parl.Model): agents' local q network for decision making.
            qmixer_model (parl.Model): A mixing network which takes local q values as input to construct a global Q network.
            double_q (bool): Double-DQN.
            gamma (float): discounted factor for reward computation.
            lr (float): learning rate.
            clip_grad_norm (None, or float): clipped value of gradients' global norm.
        """
        self.agent_model = agent_model
        self.qmixer_model = qmixer_model
        self.target_agent_model = deepcopy(self.agent_model)
        self.target_qmixer_model = deepcopy(self.qmixer_model)

        self.n_agents = self.qmixer_model.n_agents
        assert isinstance(gamma, float)
        assert isinstance(lr, float)
        self.double_q = double_q
        self.gamma = gamma
        self.lr = lr
        self.clip_grad_norm = clip_grad_norm

        self.params = list(self.agent_model.parameters())
        self.params += self.qmixer_model.parameters()
        if self.clip_grad_norm:
            clip = nn.ClipGradByGlobalNorm(clip_norm=self.clip_grad_norm)
            self.optimizer = paddle.optimizer.RMSProp(
                parameters=self.params,
                learning_rate=self.lr,
                rho=0.99,
                epsilon=1e-5,
                grad_clip=clip)
        else:
            self.optimizer = paddle.optimizer.RMSProp(
                parameters=self.params,
                learning_rate=self.lr,
                rho=0.99,
                epsilon=1e-5)

    def _init_hidden_states(self, batch_size):
        self.hidden_states = self.agent_model.init_hidden().unsqueeze(
            0).expand(shape=(batch_size, self.n_agents, -1))
        self.target_hidden_states = self.target_agent_model.init_hidden(
        ).unsqueeze(0).expand(shape=(batch_size, self.n_agents, -1))

    def predict_local_q(self, obs, hidden_state):
        return self.agent_model(obs, hidden_state)

    def learn(self, state_batch, actions_batch, reward_batch, terminated_batch,
              obs_batch, available_actions_batch, filled_batch):
        """
        Args:
            state_batch (paddle.Tensor):             (batch_size, T, state_shape)
            actions_batch (paddle.Tensor):           (batch_size, T, n_agents)
            reward_batch (paddle.Tensor):            (batch_size, T, 1)
            terminated_batch (paddle.Tensor):        (batch_size, T, 1)
            obs_batch (paddle.Tensor):               (batch_size, T, n_agents, obs_shape)
            available_actions_batch (paddle.Tensor): (batch_size, T, n_agents, n_actions)
            filled_batch (paddle.Tensor):            (batch_size, T, 1)
        Returns:
            loss (float): train loss
            td_error (float): train TD error
        """
        batch_size = state_batch.shape[0]
        episode_len = state_batch.shape[1]
        self._init_hidden_states(batch_size)
        n_actions = available_actions_batch.shape[-1]

        reward_batch = reward_batch[:, :-1, :]
        actions_batch = actions_batch[:, :-1, :]
        terminated_batch = terminated_batch[:, :-1, :]
        filled_batch = filled_batch[:, :-1, :]

        mask = (1 - filled_batch) * (1 - terminated_batch)

        local_qs = []
        target_local_qs = []
        for t in range(episode_len):
            obs = obs_batch[:, t, :, :]
            obs = obs.reshape(shape=(-1, obs_batch.shape[-1]))
            local_q, self.hidden_states = self.agent_model(
                obs, self.hidden_states)
            local_q = local_q.reshape(shape=(batch_size, self.n_agents, -1))
            local_qs.append(local_q)

            target_local_q, self.target_hidden_states = self.target_agent_model(
                obs, self.target_hidden_states)
            target_local_q = target_local_q.reshape(
                shape=(batch_size, self.n_agents, -1))
            target_local_qs.append(target_local_q)

        local_qs = paddle.stack(local_qs, axis=1)
        target_local_qs = paddle.stack(target_local_qs[1:], axis=1)

        actions_batch_one_hot = F.one_hot(actions_batch, num_classes=n_actions)
        chosen_action_local_qs = paddle.sum(
            local_qs[:, :-1, :, :] * actions_batch_one_hot, axis=-1)
        # mask unavailable actions
        target_unavailable_actions_mask = (
            available_actions_batch[:, 1:, :] == 0).cast('float32')
        target_local_qs -= 1e8 * target_unavailable_actions_mask
        if self.double_q:
            local_qs_detach = local_qs.clone().detach()
            unavailable_actions_mask = (
                available_actions_batch == 0).cast('float32')
            local_qs_detach -= 1e8 * unavailable_actions_mask
            cur_max_actions = paddle.argmax(
                local_qs_detach[:, 1:], axis=-1, keepdim=False)
            cur_max_actions_one_hot = F.one_hot(
                cur_max_actions, num_classes=n_actions)
            target_local_max_qs = paddle.sum(
                target_local_qs * cur_max_actions_one_hot, axis=-1)
        else:
            target_local_max_qs = target_local_qs.max(axis=3)

        chosen_action_global_qs = self.qmixer_model(chosen_action_local_qs,
                                                    state_batch[:, :-1, :])
        target_global_max_qs = self.target_qmixer_model(
            target_local_max_qs, state_batch[:, 1:, :])

        target = reward_batch + self.gamma * (
            1 - terminated_batch) * target_global_max_qs
        td_error = target.detach() - chosen_action_global_qs
        masked_td_error = td_error * mask
        mean_td_error = masked_td_error.sum() / mask.sum()
        loss = (masked_td_error**2).sum() / mask.sum()

        self.optimizer.clear_grad()
        loss.backward()
        self.optimizer.step()
        return loss.numpy()[0], mean_td_error.numpy()[0]

    def sync_target(self):
        self.agent_model.sync_weights_to(self.target_agent_model)
        self.qmixer_model.sync_weights_to(self.target_qmixer_model)



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
warnings.simplefilter('default')

from parl.core.fluid import layers
from copy import deepcopy
from paddle import fluid
from parl.core.fluid.algorithm import Algorithm

__all__ = ['MADDPG']

from parl.core.fluid.policy_distribution import SoftCategoricalDistribution
from parl.core.fluid.policy_distribution import SoftMultiCategoricalDistribution


def SoftPDistribution(logits, act_space):
    """Args:
            logits: the output of policy model
            act_space: action space, must be gym.spaces.Discrete or multiagent.multi_discrete.MultiDiscrete

        Return:
            instance of SoftCategoricalDistribution or SoftMultiCategoricalDistribution
    """
    # is instance of gym.spaces.Discrete
    if (hasattr(act_space, 'n')):
        return SoftCategoricalDistribution(logits)
    # is instance of multiagent.multi_discrete.MultiDiscrete
    elif (hasattr(act_space, 'num_discrete_space')):
        return SoftMultiCategoricalDistribution(logits, act_space.low,
                                                act_space.high)
    else:
        raise AssertionError("act_space must be instance of \
            gym.spaces.Discrete or multiagent.multi_discrete.MultiDiscrete")


class MADDPG(Algorithm):
    def __init__(self,
                 model,
                 agent_index=None,
                 act_space=None,
                 gamma=None,
                 tau=None,
                 lr=None,
                 actor_lr=None,
                 critic_lr=None):
        """  MADDPG algorithm
        
        Args:
            model (parl.Model): forward network of actor and critic.
                                The function get_actor_params() of model should be implemented.
            agent_index: index of agent, in multiagent env
            act_space: action_space, gym space
            gamma (float): discounted factor for reward computation.
            tau (float): decay coefficient when updating the weights of self.target_model with self.model
            lr (float): learning rate, lr will be assigned to both critic_lr and actor_lr
            critic_lr (float): learning rate of the critic model
            actor_lr (float): learning rate of the actor model
        """

        assert isinstance(agent_index, int)
        assert isinstance(act_space, list)
        assert isinstance(gamma, float)
        assert isinstance(tau, float)
        # compatible upgrade of lr
        if lr is None:
            assert isinstance(actor_lr, float)
            assert isinstance(critic_lr, float)
        else:
            assert isinstance(lr, float)
            assert actor_lr is None, 'no need to set `actor_lr` if `lr` is not None'
            assert critic_lr is None, 'no need to set `critic_lr` if `lr` is not None'
            critic_lr = lr
            actor_lr = lr
            warnings.warn(
                "the `lr` argument of `__init__` function in `parl.Algorithms.MADDPG` is deprecated \
                    since version 1.4 and will be removed in version 2.0. \
                    Recommend to use `actor_lr` and `critic_lr`. ",
                DeprecationWarning,
                stacklevel=2)
        self.agent_index = agent_index
        self.act_space = act_space
        self.gamma = gamma
        self.tau = tau
        self.lr = lr
        self.actor_lr = actor_lr
        self.critic_lr = critic_lr

        self.model = model
        self.target_model = deepcopy(model)

    def predict(self, obs):
        """ input:  
                obs: observation, shape([B] + shape of obs_n[agent_index])
            output: 
                act: action, shape([B] + shape of act_n[agent_index])
        """
        this_policy = self.model.policy(obs)
        this_action = SoftPDistribution(
            logits=this_policy,
            act_space=self.act_space[self.agent_index]).sample()
        return this_action

    def predict_next(self, obs):
        """ input:  observation, shape([B] + shape of obs_n[agent_index])
            output: action, shape([B] + shape of act_n[agent_index])
        """
        next_policy = self.target_model.policy(obs)
        next_action = SoftPDistribution(
            logits=next_policy,
            act_space=self.act_space[self.agent_index]).sample()
        return next_action

    def Q(self, obs_n, act_n):
        """ input:  
                obs_n: all agents' observation, shape([B] + shape of obs_n)
            output: 
                act_n: all agents' action, shape([B] + shape of act_n)
        """
        return self.model.value(obs_n, act_n)

    def Q_next(self, obs_n, act_n):
        """ input:  
                obs_n: all agents' observation, shape([B] + shape of obs_n)
            output: 
                act_n: all agents' action, shape([B] + shape of act_n)
        """
        return self.target_model.value(obs_n, act_n)

    def learn(self, obs_n, act_n, target_q):
        """ update actor and critic model with MADDPG algorithm
        """
        actor_cost = self._actor_learn(obs_n, act_n)
        critic_cost = self._critic_learn(obs_n, act_n, target_q)
        return critic_cost

    def _actor_learn(self, obs_n, act_n):
        i = self.agent_index
        this_policy = self.model.policy(obs_n[i])
        sample_this_action = SoftPDistribution(
            logits=this_policy,
            act_space=self.act_space[self.agent_index]).sample()

        action_input_n = act_n + []
        action_input_n[i] = sample_this_action
        eval_q = self.Q(obs_n, action_input_n)
        act_cost = layers.reduce_mean(-1.0 * eval_q)

        act_reg = layers.reduce_mean(layers.square(this_policy))

        cost = act_cost + act_reg * 1e-3

        fluid.clip.set_gradient_clip(
            clip=fluid.clip.GradientClipByNorm(clip_norm=0.5),
            param_list=self.model.get_actor_params())

        optimizer = fluid.optimizer.AdamOptimizer(self.actor_lr)
        optimizer.minimize(cost, parameter_list=self.model.get_actor_params())
        return cost

    def _critic_learn(self, obs_n, act_n, target_q):
        pred_q = self.Q(obs_n, act_n)
        cost = layers.reduce_mean(layers.square_error_cost(pred_q, target_q))

        fluid.clip.set_gradient_clip(
            clip=fluid.clip.GradientClipByNorm(clip_norm=0.5),
            param_list=self.model.get_critic_params())

        optimizer = fluid.optimizer.AdamOptimizer(self.critic_lr)
        optimizer.minimize(cost, parameter_list=self.model.get_critic_params())
        return cost

    def sync_target(self, decay=None):
        if decay is None:
            decay = 1.0 - self.tau
        self.model.sync_weights_to(self.target_model, decay=decay)



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import threading
from six.moves import queue
from parl.remote.exceptions import FutureGetRepeatedlyError, FutureFunctionError, FutureObjectEmpty


class FutureObject(object):
    def __init__(self, output_queue):
        """This class is used to encapsulate the output_queue(`queue.Queue`),
        and provides a `get` function. When calling a function of a class
        decorated by `pare.remote_class(wait=False)`, user will get a `FutureObject`
        immediately and can get the real return by calling the `get` function of 
        the `future` object.

        For example:
            ```python
            import parl

            @parl.remote_class(wait=False)
            class Actor(object):
                def __init__(self):
                    pass

                def func(self):
                    # do something
                    return 0

            parl.connect("localhost:8010")

            actors = Actor()
            future_object = actor.func() # return a `FutureObject`
            result = future_object.get()
            ```

        Args:
            output_queue(queue.Queue): queue to get the result of function calling
                                           or set_attr/get_attr
        """
        self._output_queue = output_queue
        self._already_get = False
        self.internal_lock = threading.Lock()

    def get(self, block=True, timeout=None):
        """
        Remove and return the data from the `FutureObject`. 
        If optional args block is true and timeout is None (the default), block if 
        necessary until an item is available. If timeout is a positive number, it 
        blocks at most timeout seconds and raises the `parl.remote.exceptions.FutureObjectEmpty`
        exception if no itemwas available within that time. Otherwise (block is false), return an item
        if one is immediately available, else raise the `parl.remote.exceptions.FutureObjectEmpty`
        exception (timeout is ignored in that case).
        """
        if timeout is not None:
            assert timeout > 0, "`timeout` must be a non-negative number"

        with self.internal_lock:
            if self._already_get:
                raise FutureGetRepeatedlyError()

            try:
                result = self._output_queue.get(block=block, timeout=timeout)
            except queue.Empty:
                raise FutureObjectEmpty

            if isinstance(result, FutureFunctionError):
                time.sleep(
                    0.1
                )  # waiting for another thread printing the error message
                raise result
            self._already_get = True
            return result

    def get_nowait(self):
        """Equivalent to get(block=False).
        """
        return self.get(block=False)

    def empty(self):
        """
        Return True if the `FutureObject` is empty, False otherwise.
        if empty() returns False it doesn’t guarantee that a subsequent call to get() will not block.
        """
        return self._output_queue.empty()



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import threading
from collections import defaultdict


class JobCenter(object):
    """The job center deals with everythin related to jobs.

    Attributes:
        job_pool (set): A set to store the job address of vacant cpu.
        worker_dict (dict): A dict to store connected workers.
        worker_hostname (dict): A dict to record worker hostname.
        worker_vacant_jobs (dict): Record how many vacant jobs does each
                                   worker has.
        master_ip (str): IP address of the master node.
    """

    def __init__(self, master_ip):
        self.job_pool = dict()
        self.worker_dict = {}
        self.worker_hostname = defaultdict(int)
        self.worker_vacant_jobs = {}
        self.lock = threading.Lock()
        self.master_ip = master_ip

    @property
    def cpu_num(self):
        """"Return vacant cpu number."""
        return len(self.job_pool)

    @property
    def worker_num(self):
        """Return connected worker number."""
        return len(self.worker_dict)

    def add_worker(self, worker):
        """When a new worker connects, add its hostname to worker_hostname.

        Args:
            worker (InitializedWorker): New worker with initialized jobs.
        """
        self.lock.acquire()
        self.worker_dict[worker.worker_address] = worker
        for job in worker.initialized_jobs:
            self.job_pool[job.job_address] = job

        self.worker_vacant_jobs[worker.worker_address] = len(
            worker.initialized_jobs)

        if self.master_ip and worker.worker_address.split(
                ':')[0] == self.master_ip:
            self.worker_hostname[worker.worker_address] = "Master"
            self.master_ip = None
        else:
            self.worker_hostname[worker.hostname] += 1
            self.worker_hostname[worker.worker_address] = "{}:{}".format(
                worker.hostname, self.worker_hostname[worker.hostname])
        self.lock.release()

    def drop_worker(self, worker_address):
        """Remove jobs from job_pool when a worker dies.

        Args:
            worker_address (str): the worker_address of a worker to be
                                  removed from the job center.
        """
        self.lock.acquire()
        worker = self.worker_dict[worker_address]
        for job in worker.initialized_jobs:
            if job.job_address in self.job_pool:
                self.job_pool.pop(job.job_address)
        self.worker_dict.pop(worker_address)
        self.worker_vacant_jobs.pop(worker_address)
        self.lock.release()

    def request_job(self):
        """Return a job_address when the client submits a job.

        If there is no vacant CPU in the cluster, this will return None.

        Return:
            An ``InitializedJob`` that has information about available job address.
        """
        self.lock.acquire()
        job = None
        if len(self.job_pool):
            job_address, job = self.job_pool.popitem()
            self.worker_vacant_jobs[job.worker_address] -= 1
            assert self.worker_vacant_jobs[job.worker_address] >= 0
        self.lock.release()
        return job

    def reset_job(self, job):
        """Reset a job and add the job_address to the job_pool.

        Args:
            job(``InitializedJob``): The job information of the restarted job.
        """
        self.lock.acquire()
        self.job_pool[job.job_address] = job
        self.lock.release()

    def update_job(self, killed_job_address, new_job, worker_address):
        """When worker kill an old job, it will start a new job.

        Args:
            killed_job_address (str): The job address of the killed job.
            new_job(``InitializedJob``): Information of the new job.
            worker_address (str): The worker which kills an old job.
        """
        self.lock.acquire()
        self.job_pool[new_job.job_address] = new_job

        if killed_job_address in self.job_pool:
            self.job_pool.pop(killed_job_address)

        to_del_idx = None
        for i, job in enumerate(
                self.worker_dict[worker_address].initialized_jobs):
            if job.job_address == killed_job_address:
                to_del_idx = i
                break

        del self.worker_dict[worker_address].initialized_jobs[to_del_idx]
        self.worker_dict[worker_address].initialized_jobs.append(new_job)

        if killed_job_address not in self.job_pool:
            self.worker_vacant_jobs[worker_address] += 1

        self.lock.release()

    def get_vacant_cpu(self, worker_address):
        """Return vacant cpu number of a worker."""
        return self.worker_vacant_jobs[worker_address]

    def get_total_cpu(self, worker_address):
        """Return total cpu number of a worker."""
        return len(self.worker_dict[worker_address].initialized_jobs)

    def get_hostname(self, worker_address):
        """Return the hostname of a worker."""
        return self.worker_hostname[worker_address]



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import numpy as np
import torch
import torch.nn.functional as F
from copy import deepcopy
import parl

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

__all__ = ['TD3']


class TD3(parl.Algorithm):
    def __init__(
            self,
            model,
            gamma=None,
            tau=None,
            actor_lr=None,
            critic_lr=None,
            policy_noise=0.2,  # Noise added to target policy during critic update
            noise_clip=0.5,  # Range to clip target policy noise
            policy_freq=2):  # Frequency of delayed policy updates
        assert isinstance(gamma, float)
        assert isinstance(tau, float)
        assert isinstance(actor_lr, float)
        assert isinstance(critic_lr, float)

        self.gamma = gamma
        self.tau = tau
        self.actor_lr = actor_lr
        self.critic_lr = critic_lr
        self.policy_noise = policy_noise
        self.noise_clip = noise_clip
        self.policy_freq = policy_freq

        self.model = model.to(device)
        self.target_model = deepcopy(model).to(device)
        self.actor_optimizer = torch.optim.Adam(
            self.model.get_actor_params(), lr=actor_lr)
        self.critic_optimizer = torch.optim.Adam(
            self.model.get_critic_params(), lr=critic_lr)

        self.total_it = 0

    def predict(self, obs):
        return self.model.policy(obs)

    def learn(self, obs, action, reward, next_obs, terminal):
        self.total_it += 1
        self._critic_learn(obs, action, reward, next_obs, terminal)
        if self.total_it % self.policy_freq == 0:
            self._actor_learn(obs)

    def _critic_learn(self, obs, action, reward, next_obs, terminal):
        with torch.no_grad():
            noise = (torch.randn_like(action) * self.policy_noise).clamp(
                -self.noise_clip, self.noise_clip)

            next_action = (self.target_model.policy(next_obs) + noise).clamp(
                -1, 1)

            target_Q1, target_Q2 = self.target_model.value(
                next_obs, next_action)
            target_Q = torch.min(target_Q1, target_Q2)
            target_Q = reward + (1 - terminal) * self.gamma * target_Q

        current_Q1, current_Q2 = self.model.value(obs, action)

        critic_loss = F.mse_loss(current_Q1, target_Q) + F.mse_loss(
            current_Q2, target_Q)

        self.critic_optimizer.zero_grad()
        critic_loss.backward()
        self.critic_optimizer.step()

    def _actor_learn(self, obs):
        actor_loss = -self.model.Q1(obs, self.model.policy(obs)).mean()

        self.actor_optimizer.zero_grad()
        actor_loss.backward()
        self.actor_optimizer.step()

        self.sync_target()

    def sync_target(self, decay=None):
        if decay is None:
            decay = 1.0 - self.tau
        for param, target_param in zip(self.model.parameters(),
                                       self.target_model.parameters()):
            target_param.data.copy_((1 - decay) * param.data +
                                    decay * target_param.data)



#   Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import paddle
import paddle.nn as nn
import collections
import numpy as np

from parl.core.model_base import ModelBase
from parl.utils import machine_info

__all__ = ['Model']


class Model(nn.Layer, ModelBase):
    """
    | `alias`: ``parl.Model``
    | `alias`: ``parl.core.paddle.agent.Model``

    | ``Model`` is a base class of PARL for the neural network. 
    A ``Model`` is usually a policy or Q-value function, which predicts 
    an action or an estimate according to the environmental observation.

    | To use the ``PaddlePaddle2.0`` backend model, user needs to call 
    ``super(Model, self).__init__()`` at the beginning of ``__init__`` 
    function.

    | ``Model`` supports duplicating a ``Model`` instance in a pythonic way:

    | ``copied_model = copy.deepcopy(model)``

    Example:

    .. code-block:: python

        import parl
        import paddle.nn as nn

        class Policy(parl.Model):
            def __init__(self):
                super(Policy, self).__init__()
                self.fc = nn.Linear(input_dim=100, output_dim=32)

            def policy(self, obs):
                out = self.fc(obs)
                return out
               
        policy = Policy() 
        copied_policy = copy.deepcopy(policy)

    Attributes:
        model_id(str): each model instance has its unique model_id.

    Public Functions:
        - ``sync_weights_to``: synchronize parameters of the current model 
        to another model.
        - ``get_weights``: return a list containing all the parameters of 
        the current model.
        - ``set_weights``: copy parameters from ``set_weights()`` to the model.
        - ``forward``: define the computations of a neural network. **Should** 
        be overridden by all subclasses.
    """

    def __init___(self):
        super(Model, self).__init__()

    def sync_weights_to(self, target_model, decay=0.0):
        """Synchronize parameters of current model to another model.

        target_model_weights = decay * target_model_weights 
                                    + (1 - decay) * current_model_weights

        Args:
            target_model (`parl.Model`): an instance of ``Model`` that has 
                the same neural network architecture as the current model.
            decay (float):  the rate of decline in copying parameters. 
                0 if no parameters decay when synchronizing the parameters.

        Example:

        .. code-block:: python

            import copy
            # create a model that has the same neural network structures.
            target_model = copy.deepcopy(model)

            # after initilizing the parameters ...
            model.sync_weights_to(target_mdodel)

        Note:
            Before calling ``sync_weights_to``, parameters of the model must 
            have been initialized.
        """
        assert not target_model is self, "cannot copy between identical model"
        assert isinstance(target_model, Model)
        assert self.__class__.__name__ == target_model.__class__.__name__, \
            "must be the same class for params syncing!"
        assert (decay >= 0 and decay <= 1)

        target_vars = dict(target_model.named_parameters())
        for name, var in self.named_parameters():
            target_data = decay * target_vars[name] + (1 - decay) * var
            target_vars[name] = target_data
        target_model.set_state_dict(target_vars)

    def get_weights(self):
        """Returns a Python dict containing parameters of current model.

        Returns: 
            a Python dict containing the parameters of current model.
        """
        weights = self.state_dict()
        for key in weights.keys():
            weights[key] = weights[key].numpy()
        return weights

    def set_weights(self, weights):
        """Copy parameters from ``set_weights()`` to the model.
        
        Args:
            weights (dict): a Python dict containing the parameters.
        """
        old_weights = self.state_dict()
        assert len(old_weights) == len(
            weights), '{} params are expected, but got {}'.format(
                len(old_weights), len(weights))
        new_weights = collections.OrderedDict()
        for key in old_weights.keys():
            assert key in weights, 'key: {} is expected to be in weights.'.format(
                key)
            assert old_weights[key].shape == list(
                weights[key].shape
            ), 'key \'{}\' expects the data with shape {}, but gets {}'.format(
                key, old_weights[key].shape, list(weights[key].shape))
            new_weights[key] = paddle.to_tensor(weights[key])
        self.set_state_dict(new_weights)



#   Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import paddle
from parl.core.agent_base import AgentBase
from parl.core.paddle.algorithm import Algorithm
from parl.utils import machine_info, get_gpu_count

__all__ = ['Agent']


class Agent(AgentBase):
    """
    | `alias`: ``parl.Agent``
    | `alias`: ``parl.core.paddle.agent.Agent``

    | Agent is one of the three basic classes of PARL.

    | It is responsible for interacting with the environment and collecting 
    data for training the policy.
    | To implement a customized ``Agent``, users can:

      .. code-block:: python

        import parl

        class MyAgent(parl.Agent):
            def __init__(self, algorithm, act_dim):
                super(MyAgent, self).__init__(algorithm)
                self.act_dim = act_dim

    Attributes:
        alg (parl.algorithm): algorithm of this agent.
        place: can automatically specify device when creating a tensor.

    Public Functions:
        - ``sample``: return a noisy action to perform exploration according to the policy.
        - ``predict``: return an action given current observation.
        - ``learn``: update the parameters of self.alg using the `learn_program` defined in `build_program()`.
        - ``save``: save parameters of the ``agent`` to a given path.
        - ``restore``: restore previous saved parameters from a given path.

    Todo:
        - allow users to get parameters of a specified model by specifying the model's name in ``get_weights()``.

    """

    def __init__(self, algorithm):
        """

        Args:
            algorithm (parl.Algorithm): an instance of `parl.Algorithm`. This algorithm is then passed to `self.alg`.
        """

        assert isinstance(algorithm, Algorithm)
        super(Agent, self).__init__(algorithm)

        gpu_count = get_gpu_count()
        if gpu_count > 0:
            self.place = paddle.CUDAPlace(0)
        else:
            self.place = paddle.CPUPlace()

    def learn(self, *args, **kwargs):
        """The training interface for ``Agent``.
        """
        raise NotImplementedError

    def predict(self, *args, **kwargs):
        """Predict an action when given the observation of the environment.
        """
        raise NotImplementedError

    def sample(self, *args, **kwargs):
        """Return an action with noise when given the observation of the environment.

        In general, this function is used in train process as noise is added to the action to preform exploration.

        """
        raise NotImplementedError

    def save(self, save_path, model=None):
        """Save parameters.

        Args:
            save_path(str): where to save the parameters.
            model(parl.Model): model that describes the neural network structure. If None, will use self.alg.model.

        Raises:
            ValueError: if program is None and self.learn_program does not exist.

        Example:

        .. code-block:: python

            agent = AtariAgent()
            agent.save('./model_dir')

        """
        if model is None:
            model = self.alg.model
        paddle.save(model.state_dict(), save_path)

    def restore(self, save_path, model=None):
        """Restore previously saved parameters.
        This method requires a program that describes the network structure.
        The save_path argument is typically a value previously passed to ``save_params()``.

        Args:
            save_path(str): path where parameters were previously saved.
            model(parl.Model): model that describes the neural network structure. If None, will use self.alg.model.

        Raises:
            ValueError: if program is None and self.learn_program does not exist.

        Example:

        .. code-block:: python

            agent = AtariAgent()
            agent.save('./model_dir')
            agent.restore('./model_dir')

        """
        if model is None:
            model = self.alg.model
        param_dict = paddle.load(save_path)
        model.set_state_dict(param_dict)



#   Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gym
import numpy as np
from parl.utils import logger


class ActionMappingWrapper(gym.Wrapper):
    def __init__(self, env):
        """Map action space [-1, 1] of model output to new action space
        [low_bound, high_bound].
        """
        gym.Wrapper.__init__(self, env)
        assert isinstance(self.env.action_space, gym.spaces.Box)
        self.low_bound = self.env.action_space.low[0]
        self.high_bound = self.env.action_space.high[0]
        assert self.high_bound > self.low_bound
        if hasattr(env, '_max_episode_steps'):
            self._max_episode_steps = int(self.env._max_episode_steps)

    def reset(self, **kwargs):
        return self.env.reset(**kwargs)

    def step(self, model_output_act):
        """
        Args:
            model_output_act(np.array): The values must be in in [-1, 1].
        """
        assert np.all(((model_output_act<=1.0 + 1e-3), (model_output_act>=-1.0 - 1e-3))), \
            'the action should be in range [-1.0, 1.0]'
        mapped_action = self.low_bound + (model_output_act - (-1.0)) * (
            (self.high_bound - self.low_bound) / 2.0)
        mapped_action = np.clip(mapped_action, self.low_bound, self.high_bound)

        return self.env.step(mapped_action)



#   Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from parl.core.algorithm_base import AlgorithmBase
from parl.core.paddle.model import Model

__all__ = ['Algorithm']


class Algorithm(AlgorithmBase):
    """
    | `alias`: ``parl.Algorithm``
    | `alias`: ``parl.core.fluid.algorithm.Algorithm``

    | ``Algorithm`` defines the way how to update the parameters of 
    the ``Model``. This is where we define loss functions and the 
    optimizer of the neural network. An ``Algorithm`` has at least 
    a model.

    | PARL has implemented various algorithms(DQN/DDPG/PPO/A3C/IMPALA) that 
    can be reused quickly, which can be accessed with ``parl.algorithms``.

    Example:

    .. code-block:: python

        import parl

        model = Model()
        dqn = parl.algorithms.DQN(model, lr=1e-3)

    Attributes:
        model(``parl.Model``): a neural network that represents a policy 
        or a Q-value function.

    Pulic Functions:
        - ``get_weights``: return a Python dictionary containing parameters 
        of the current model.
        - ``set_weights``: copy parameters from ``get_weights()`` to the model.
        - ``sample``: return a noisy action to perform exploration according 
        to the policy.
        - ``predict``: return an action given current observation.
        - ``learn``: define the loss function and create an optimizer to 
        minized the loss.
    """

    def __init__(self, model=None):
        """
        Args:
            model(``parl.Model``): a neural network that represents a policy or a Q-value function.
        """
        assert isinstance(model, Model)
        self.model = model

    def learn(self, *args, **kwargs):
        """ Define the loss function and create an optimizer to minize the loss.
        """
        raise NotImplementedError

    def predict(self, *args, **kwargs):
        """ Refine the predicting process, e.g,. use the policy model to predict actions.
        """
        raise NotImplementedError

    def sample(self, *args, **kwargs):
        """ Define the sampling process. This function returns an action with noise to perform exploration.
        """
        raise NotImplementedError

    def get_weights(self):
        """ Get weights of self.model.

        Returns:
            weights (dict): a Python dict containing the parameters of self.model.
        """
        return self.model.get_weights()

    def set_weights(self, params):
        """ Set weights from ``get_weights`` to the model.

        Args:
            weights (dict): a Python dict containing the parameters of self.model.
        """
        self.model.set_weights(params)



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
warnings.simplefilter('default')

import os
import torch

from parl.core.agent_base import AgentBase
from parl.core.torch.algorithm import Algorithm
from parl.utils import machine_info

__all__ = ['Agent']
torch.set_num_threads(1)


class Agent(AgentBase):
    """
    | `alias`: ``parl.Agent``
    | `alias`: ``parl.core.torch.agent.Agent``

    | Agent is one of the three basic classes of PARL.

    | It is responsible for interacting with the environment and collecting data for training the policy.
    | To implement a customized ``Agent``, users can:

      .. code-block:: python

        import parl

        class MyAgent(parl.Agent):
            def __init__(self, algorithm, act_dim):
                super(MyAgent, self).__init__(algorithm)
                self.act_dim = act_dim

    Attributes:
        device (torch.device): select GPU/CPU to be used.
        alg (parl.Algorithm): algorithm of this agent.

    Public Functions:
        - ``sample``: return a noisy action to perform exploration according to the policy.
        - ``predict``: return an estimate Q function given current observation.
        - ``learn``: update the parameters of self.alg.
        - ``save``: save parameters of the ``agent`` to a given path.
        - ``restore``: restore previous saved parameters from a given path.

    Todo:
        - allow users to get parameters of a specified model by specifying the model's name in ``get_weights()``.
    """

    def __init__(self, algorithm):
        """.

        Args:
            algorithm (parl.Algorithm): an instance of `parl.Algorithm`. This algorithm is then passed to `self.alg`.
            device (torch.device): specify which GPU/CPU to be used.
        """

        assert isinstance(algorithm, Algorithm)
        super(Agent, self).__init__(algorithm)

    def learn(self, *args, **kwargs):
        """The training interface for ``Agent``.

        It is often used in the training stage.
        """
        raise NotImplementedError

    def predict(self, *args, **kwargs):
        """Predict an estimated Q value when given the observation of the environment.

        It is often used in the evaluation stage.
        """
        raise NotImplementedError

    def sample(self, *args, **kwargs):
        """Return an action with noise when given the observation of the environment.

        In general, this function is used in train process as noise is added to the action to preform exploration.

        """
        raise NotImplementedError

    def save(self, save_path, model=None):
        """Save parameters.

        Args:
            save_path(str): where to save the parameters.
            model(parl.Model): model that describes the neural network structure. If None, will use self.alg.model.

        Raises:
            ValueError: if model is None and self.alg.model does not exist.

        Example:

        .. code-block:: python

            agent = AtariAgent()
            agent.save('./model.ckpt')

        """
        if model is None:
            model = self.alg.model
        sep = os.sep
        dirname = sep.join(save_path.split(sep)[:-1])
        if dirname != '' and not os.path.exists(dirname):
            os.makedirs(dirname)
        torch.save(model.state_dict(), save_path)

    def restore(self, save_path, model=None, map_location=None):
        """Restore previously saved parameters.
        This method requires a model that describes the network structure.
        The save_path argument is typically a value previously passed to ``save()``.

        Args:
            save_path(str): path where parameters were previously saved.
            model(parl.Model): model that describes the neural network structure. If None, will use self.alg.model.
            map_location: a function, torch.device, string or a dict specifying how to remap storage locations

        Raises:
            ValueError: if model is None and self.alg does not exist.

        Example:

        .. code-block:: python

            agent = AtariAgent()
            agent.save('./model.ckpt')
            agent.restore('./model.ckpt')
            
            agent.restore('./model.ckpt', map_location=torch.device('cpu')) # load gpu-trained model in cpu machine
        """

        if model is None:
            model = self.alg.model
        checkpoint = torch.load(save_path, map_location=map_location)
        model.load_state_dict(checkpoint)



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import torch.optim as optim
import parl
from torch.distributions import Categorical

__all__ = ['PolicyGradient']


class PolicyGradient(parl.Algorithm):
    def __init__(self, model, lr):
        """Policy gradient algorithm

        Args:
            model (parl.Model): model defining forward network of policy.
            lr (float): learning rate.

        """
        assert isinstance(lr, float)

        self.model = model
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.model.to(device)
        self.optimizer = optim.Adam(self.model.parameters(), lr=lr)

    def predict(self, obs):
        """Predict the probability of actions

        Args:
            obs (torch.tensor): shape of (obs_dim,)

        Returns:
            prob (torch.tensor): shape of (action_dim,)
        """
        prob = self.model(obs)
        return prob

    def learn(self, obs, action, reward):
        """Update model with policy gradient algorithm

        Args:
            obs (torch.tensor): shape of (batch_size, obs_dim)
            action (torch.tensor): shape of (batch_size, 1)
            reward (torch.tensor): shape of (batch_size, 1)

        Returns:
            loss (torch.tensor): shape of (1)

        """
        prob = self.model(obs)

        log_prob = Categorical(prob).log_prob(action)

        loss = torch.mean(-1 * log_prob * reward)

        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()
        return loss



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six
from collections import defaultdict

__all__ = ['VectorEnv']


class VectorEnv(object):
    """ vector of envs to support vector reset and vector step.
    `vector_step` api will automatically reset envs which are done.
    """

    def __init__(self, envs):
        """
        Args:
            envs: List of env
        """
        self.envs = envs
        self.envs_num = len(envs)

    def reset(self):
        """
        Returns:
            List of obs
        """
        return [env.reset() for env in self.envs]

    def step(self, actions):
        """
        Args:
            actions: List or array of action

        Returns:
            obs_batch: List of next obs of envs
            reward_batch: List of return reward of envs 
            done_batch: List of done of envs 
            info_batch: List of info of envs 
        """
        obs_batch, reward_batch, done_batch, info_batch = [], [], [], []
        for env_id in six.moves.range(self.envs_num):
            obs, reward, done, info = self.envs[env_id].step(actions[env_id])

            if done:
                obs = self.envs[env_id].reset()

            obs_batch.append(obs)
            reward_batch.append(reward)
            done_batch.append(done)
            info_batch.append(info)
        return obs_batch, reward_batch, done_batch, info_batch



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Wrappers for fluid.layers. It helps to easily share parameters between layers.

NOTE:
    We only encapsulated some of layers with parameters in the fluid, which are frequently used in RL scene.
    If you need use some layers with parameters that are not encapsulated, please submit an issue
    or PR.

Here is an example:
    ```python
    from parl import layers

    class MLPModel(Model):
        def __init__(self):
            self.fc = layers.fc(size=64) # automatically create parameters names "fc_0.w" and "fc_0.b"

        def policy1(self, obs):
            out = self.fc(obs) # Really create parameters with parameters names "fc_0.w" and "fc_0.b"

        def policy2(self, obs):
            out = self.fc(obs) # Reusing parameters
    ```
"""

import inspect
import paddle.fluid.layers as layers
import paddle.fluid.unique_name as unique_name
import paddle.fluid as fluid
import six
from copy import deepcopy
from paddle.fluid.executor import _fetch_var
from paddle.fluid.framework import Variable
from paddle.fluid.layers import *
from paddle.fluid.param_attr import ParamAttr
from parl.core.fluid.layers.attr_holder import AttrHolder
from parl.utils import get_fluid_version


def update_attr_name(name, default_name, attr, is_bias):
    """
    Update the name in an attribute
    1. If the user provides a name, then generate the candidate name using the
       provided name;
    2. else generate the candidate name using the default name (which should be
       the name of the layer wrapper).
    3. After obtaining the candidate name, if the attr is False, then we return False;
    4. if the attr is None or attr.name is None, then we set the attr's name as the candidate name;
    5. else we ignore the candidate name and do nothing.
    """

    def check_or_replace_name(name, attr):
        ## if this para is not used
        if attr == False:
            return False

        if attr is None:
            return ParamAttr(name=name)

        if attr.name is None:
            attr.name = name
        return attr

    attr = ParamAttr._to_attr(attr)
    name = (default_name if name is None else name)
    suffix = "b" if is_bias else "w"
    new_name = unique_name.generate(name + "." + suffix)
    return check_or_replace_name(new_name, attr)


class LayerFunc(object):
    def __init__(self, attr_holder):
        self.attr_holder = attr_holder

    def __deepcopy__(self, memo):
        cls = self.__class__
        ## __new__ won't init the class, we need to do that ourselves
        copied = cls.__new__(cls)
        ## record in the memo that self has been copied to avoid recursive copying
        memo[id(self)] = copied

        ## first copy all content
        for k, v in six.iteritems(self.__dict__):
            setattr(copied, k, deepcopy(v, memo))

        ## then we need to create new para names for param_attr in self.attr_holder
        def create_new_para_name(attr):
            if attr:
                assert attr.name, "attr should have a name already!"
                name_key = 'PARL_target_' + attr.name
                attr.name = unique_name.generate(name_key)

        for attr in copied.attr_holder.tolist():
            create_new_para_name(attr)

        ## We require the user to sync the parameter values later, because
        ## this deepcopy is supposed to be called only before the startup
        ## program. This function will cause the computation graph change, so
        ## it cannot be called during the execution.
        return copied

    @property
    def param_name(self):
        if self.attr_holder.param_attr:
            return self.attr_holder.param_attr.name
        else:
            return None

    @property
    def bias_name(self):
        if self.attr_holder.bias_attr:
            return self.attr_holder.bias_attr.name
        else:
            return None

    @property
    def all_params_names(self):
        params_names = []
        for attr in self.attr_holder.tolist():
            if attr:
                params_names.append(attr.name)
        return params_names


class GRULayerFunc(LayerFunc):
    '''LayerFunc for GRUCell
    '''

    def __init__(self, attr_holder):
        super(GRULayerFunc, self).__init__(attr_holder)

    def __deepcopy__(self, memo):
        cls = self.__class__
        ## __new__ won't init the class, we need to do that ourselves
        copied = cls.__new__(cls)
        ## record in the memo that self has been copied to avoid recursive copying
        memo[id(self)] = copied

        ## first copy all content
        for k, v in six.iteritems(self.__dict__):
            setattr(copied, k, deepcopy(v, memo))

        ## then we need to create new para names for param_attr in self.attr_holder
        self.param_prefix = None
        self.bias_prefix = None

        def create_new_para_name(attr):
            if attr:
                assert attr.name, "attr should have a name already!"
                name_key = 'PARL_target_' + attr.name
                if '_gate' in name_key and 'param' in name_key:
                    if self.param_prefix is not None:
                        name_key = self.param_prefix
                    else:
                        name_key = name_key.replace('_gate', '')
                        name_key = unique_name.generate(name_key)
                        self.param_prefix = name_key
                    name_key += '_gate'
                elif '_gate' in name_key and 'bias' in name_key:
                    if self.bias_prefix is not None:
                        name_key = self.bias_prefix
                    else:
                        name_key = name_key.replace('_gate', '')
                        name_key = unique_name.generate(name_key)
                        self.bias_prefix = name_key
                    name_key += '_gate'
                elif '_candidate' in name_key and 'param' in name_key:
                    if self.param_prefix is not None:
                        name_key = self.param_prefix
                    else:
                        name_key = name_key.replace('_candidate', '')
                        name_key = unique_name.generate(name_key)
                        self.param_prefix = name_key
                    name_key += '_candidate'
                elif '_candidate' in name_key and 'bias' in name_key:
                    if self.bias_prefix is not None:
                        name_key = self.bias_prefix
                    else:
                        name_key = name_key.replace('_candidate', '')
                        name_key = unique_name.generate(name_key)
                        self.bias_prefix = name_key
                    name_key += '_candidate'
                else:
                    name_key = unique_name.generate(name_key)
                attr.name = name_key

        for attr in copied.attr_holder.tolist():
            create_new_para_name(attr)

        ## We require the user to sync the parameter values later, because
        ## this deepcopy is supposed to be called only before the startup
        ## program. This function will cause the computation graph change, so
        ## it cannot be called during the execution.
        return copied

    @property
    def param_candidate_name(self):
        if self.attr_holder.param_candidate_attr:
            return self.attr_holder.param_candidate_attr.name
        else:
            return None

    @property
    def param_gate_name(self):
        if self.attr_holder.param_gate_attr:
            return self.attr_holder.param_gate_attr.name
        else:
            return None

    @property
    def bias_candidate_name(self):
        if self.attr_holder.bias_candidate_attr:
            return self.attr_holder.bias_candidate_attr.name
        else:
            return None

    @property
    def bias_gate_name(self):
        if self.attr_holder.bias_gate_attr:
            return self.attr_holder.bias_gate_attr.name
        else:
            return None


def fc(size,
       num_flatten_dims=1,
       param_attr=None,
       bias_attr=None,
       act=None,
       name=None):
    """
    Return a function that creates a paddle.fluid.layers.fc.
    """
    default_name = "fc"
    param_attr = update_attr_name(name, default_name, param_attr, False)
    bias_attr = update_attr_name(name, default_name, bias_attr, True)

    class FC_(LayerFunc):
        def __init__(self):
            super(FC_, self).__init__(
                AttrHolder(param_attr=param_attr, bias_attr=bias_attr))

        def __call__(self, input):
            return layers.fc(
                input=input,
                size=size,
                num_flatten_dims=num_flatten_dims,
                param_attr=self.attr_holder.param_attr,
                bias_attr=self.attr_holder.bias_attr,
                act=act)

    return FC_()


def GRUCell(hidden_size,
            gate_activation=None,
            activation=None,
            dtype="float32",
            name=None):
    """
    Return a function that creates a paddle.fluid.layers.GRUCell.
    """
    default_name = 'GRUCell'
    if name is None:
        name = default_name
    w_name = unique_name.generate(name + '_param')
    b_name = unique_name.generate(name + '_bias')

    param_gate_attr = ParamAttr(name=w_name + '_gate')
    param_candidate_attr = ParamAttr(name=w_name + '_candidate')
    bias_gate_attr = ParamAttr(name=b_name + '_gate')
    bias_candidate_attr = ParamAttr(name=b_name + '_candidate')

    class GRUCell_(GRULayerFunc):
        def __init__(self):
            super(GRUCell_, self).__init__(
                AttrHolder(
                    param_gate_attr=param_gate_attr,
                    param_candidate_attr=param_candidate_attr,
                    bias_gate_attr=bias_gate_attr,
                    bias_candidate_attr=bias_candidate_attr))

        def __call__(self, inputs, states):
            # get the prefix name of the parameters first
            param_name_idx = self.attr_holder.param_gate_attr.name.find(
                '_gate')
            bias_name_idx = self.attr_holder.bias_gate_attr.name.find('_gate')
            assert param_name_idx is not -1 and bias_name_idx is not -1
            param_name = self.attr_holder.param_gate_attr.name[:param_name_idx]
            bias_name = self.attr_holder.bias_gate_attr.name[:bias_name_idx]
            param_attr = ParamAttr(name=param_name)
            bias_attr = ParamAttr(name=bias_name)
            gru_cell = layers.GRUCell(
                hidden_size=hidden_size,
                param_attr=param_attr,
                bias_attr=bias_attr,
                gate_activation=gate_activation,
                activation=activation,
                dtype=dtype)
            return gru_cell(inputs, states)

    return GRUCell_()


def embedding(size,
              is_sparse=False,
              is_distributed=False,
              padding_idx=None,
              param_attr=None,
              dtype="float32",
              name=None):
    """
    Return a function that creates a paddle.fluid.layers.embedding.
    """
    param_attr = update_attr_name(name, "embedding", param_attr, False)

    class Embedding_(LayerFunc):
        def __init__(self):
            super(Embedding_, self).__init__(AttrHolder(param_attr=param_attr))

        def __call__(self, input):
            return layers.embedding(
                input=input,
                size=size,
                is_sparse=is_sparse,
                is_distributed=is_distributed,
                padding_idx=padding_idx,
                param_attr=self.attr_holder.param_attr,
                dtype=dtype)

    return Embedding_()


def dynamic_lstm(size,
                 param_attr=None,
                 bias_attr=None,
                 use_peepholes=True,
                 is_reverse=False,
                 gate_activation="sigmoid",
                 cell_activation="tanh",
                 candidate_activation="tanh",
                 dtype="float32",
                 name=None):
    """
    Return a function that creates a paddle.fluid.layers.dynamic_lstm.
    """
    default_name = "dynamic_lstm"
    param_attr = update_attr_name(name, default_name, param_attr, False)
    bias_attr = update_attr_name(name, default_name, bias_attr, True)

    class DynamicLstm_(LayerFunc):
        def __init__(self):
            super(DynamicLstm_, self).__init__(
                AttrHolder(param_attr=param_attr, bias_attr=bias_attr))

        def __call__(self, input, h_0=None, c_0=None):
            return layers.dynamic_lstm(
                input=input,
                h_0=h_0,
                c_0=c_0,
                size=size,
                param_attr=self.attr_holder.param_attr,
                bias_attr=self.attr_holder.bias_attr,
                use_peepholes=use_peepholes,
                is_reverse=is_reverse,
                gate_activation=gate_activation,
                cell_activation=cell_activation,
                candidate_activation=candidate_activation,
                dtype=dtype)

    return DynamicLstm_()


def dynamic_lstmp(size,
                  proj_size,
                  param_attr=None,
                  bias_attr=None,
                  use_peepholes=True,
                  is_reverse=False,
                  gate_activation='sigmoid',
                  cell_activation='tanh',
                  candidate_activation='tanh',
                  proj_activation='tanh',
                  dtype='float32',
                  name=None,
                  cell_clip=None,
                  proj_clip=None):
    """
    Return a function that creates a paddle.fluid.layers.dynamic_lstmp.
    """
    default_name = "dynamic_lstmp"
    param_attr = update_attr_name(name, default_name, param_attr, False)
    bias_attr = update_attr_name(name, default_name, bias_attr, True)

    class DynamicLstmp_(LayerFunc):
        def __init__(self):
            super(DynamicLstmp_, self).__init__(
                AttrHolder(param_attr=param_attr, bias_attr=bias_attr))

        def __call__(self, input):
            return layers.dynamic_lstmp(
                input=input,
                size=size,
                proj_size=proj_size,
                param_attr=self.attr_holder.param_attr,
                bias_attr=self.attr_holder.bias_attr,
                use_peepholes=use_peepholes,
                is_reverse=is_reverse,
                gate_activation=gate_activation,
                cell_activation=cell_activation,
                candidate_activation=candidate_activation,
                proj_activation=proj_activation,
                dtype=dtype,
                cell_clip=cell_clip,
                proj_clip=proj_clip)

    return DynamicLstmp_()


def dynamic_gru(size,
                param_attr=None,
                bias_attr=None,
                is_reverse=False,
                gate_activation='sigmoid',
                candidate_activation='tanh',
                origin_mode=False,
                name=None):
    """
    Return a function that creates a paddle.fluid.layers.dynamic_gru.
    """
    default_name = "dynamic_gru"
    param_attr = update_attr_name(name, default_name, param_attr, False)
    bias_attr = update_attr_name(name, default_name, bias_attr, True)

    class DynamicGru_(LayerFunc):
        def __init__(self):
            super(DynamicGru_, self).__init__(
                AttrHolder(param_attr=param_attr, bias_attr=bias_attr))

        def __call__(self, input, h_0=None):
            return layers.dynamic_gru(
                input=input,
                size=size,
                param_attr=self.attr_holder.param_attr,
                bias_attr=self.attr_holder.bias_attr,
                is_reverse=is_reverse,
                gate_activation=gate_activation,
                candidate_activation=candidate_activation,
                h_0=h_0,
                origin_mode=origin_mode)

    return DynamicGru_()


def gru_unit(**kwargs):
    """
    We cannot pass param_attr or bias_attr to paddle.fluid.layers.gru_unit yet.
    """
    raise NotImplementedError()


def linear_chain_crf(**kwargs):
    raise NotImplementedError()


def crf_decoding(**kwargs):
    raise NotImplementedError()


def sequence_conv(num_filters,
                  filter_size=3,
                  filter_stride=1,
                  padding=None,
                  bias_attr=None,
                  param_attr=None,
                  act=None,
                  name=None):
    """
    Return a function that creates a paddle.fluid.layers.sequence_conv.
    """
    default_name = "sequence_conv"
    param_attr = update_attr_name(name, default_name, param_attr, False)
    bias_attr = update_attr_name(name, default_name, bias_attr, True)

    class SequenceConv_(LayerFunc):
        def __init__(self):
            super(SequenceConv_, self).__init__(
                AttrHolder(param_attr=param_attr, bias_attr=bias_attr))

        def __call__(self, input):
            return layers.sequence_conv(
                input=input,
                num_filters=num_filters,
                filter_size=filter_size,
                filter_stride=filter_stride,
                padding=padding,
                bias_attr=self.attr_holder.bias_attr,
                param_attr=self.attr_holder.param_attr,
                act=act)

    return SequenceConv_()


def conv2d(num_filters,
           filter_size,
           stride=1,
           padding=0,
           dilation=1,
           groups=None,
           param_attr=None,
           bias_attr=None,
           use_cudnn=True,
           act=None,
           name=None):
    """
    Return a function that creates a paddle.fluid.layers.conv2d.
    """
    default_name = "conv2d"
    param_attr = update_attr_name(name, default_name, param_attr, False)
    bias_attr = update_attr_name(name, default_name, bias_attr, True)

    class Conv2D_(LayerFunc):
        def __init__(self):
            super(Conv2D_, self).__init__(
                AttrHolder(param_attr=param_attr, bias_attr=bias_attr))

        def __call__(self, input):
            return layers.conv2d(
                input=input,
                num_filters=num_filters,
                filter_size=filter_size,
                stride=stride,
                padding=padding,
                dilation=dilation,
                groups=groups,
                param_attr=self.attr_holder.param_attr,
                bias_attr=self.attr_holder.bias_attr,
                use_cudnn=use_cudnn,
                act=act)

    return Conv2D_()


def conv2d_transpose(num_filters,
                     output_size=None,
                     filter_size=None,
                     padding=0,
                     stride=1,
                     dilation=1,
                     groups=None,
                     param_attr=None,
                     bias_attr=None,
                     use_cudnn=True,
                     act=None,
                     name=None):
    """
    Return a function that creates a paddle.fluid.layers.conv2d_transpose.
    """
    default_name = "conv2d_transpose"
    param_attr = update_attr_name(name, default_name, param_attr, False)
    bias_attr = update_attr_name(name, default_name, bias_attr, True)

    class Conv2DTranspose_(LayerFunc):
        def __init__(self):
            super(Conv2DTranspose_, self).__init__(
                AttrHolder(param_attr=param_attr, bias_attr=bias_attr))

        def __call__(self, input):
            return layers.conv2d_transpose(
                input=input,
                num_filters=num_filters,
                output_size=output_size,
                filter_size=filter_size,
                padding=padding,
                stride=stride,
                dilation=dilation,
                groups=groups,
                param_attr=self.attr_holder.param_attr,
                bias_attr=self.attr_holder.bias_attr,
                use_cudnn=use_cudnn,
                act=act)

    return Conv2DTranspose_()


def lstm_unit(forget_bias=0.0, param_attr=None, bias_attr=None, name=None):
    """
    Return a function that creates a paddle.fluid.layers.lstm_unit.
    """
    default_name = "lstm_unit"
    param_attr = update_attr_name(name, default_name, param_attr, False)
    bias_attr = update_attr_name(name, default_name, bias_attr, True)

    class LstmUnit_(LayerFunc):
        def __init__(self):
            super(LstmUnit_, self).__init__(
                AttrHolder(param_attr=param_attr, bias_attr=bias_attr))

        def __call__(self, x_t, hidden_t_prev, cell_t_prev):
            return layers.lstm_unit(
                x_t=x_t,
                hidden_t_prev=hidden_t_prev,
                cell_t_prev=cell_t_prev,
                forget_bias=forget_bias,
                param_attr=self.attr_holder.param_attr,
                bias_attr=self.attr_holder.bias_attr)

    return LstmUnit_()


def nce(**kwargs):
    raise NotImplementedError()


def row_conv(future_context_size, param_attr=None, act=None, name=None):
    """
    Return a function that creates a paddle.fluid.layers.row_conv.
    """
    param_attr = update_attr_name(name, "row_conv", param_attr, False)

    class RowConv_(LayerFunc):
        def __init__(self):
            super(RowConv_, self).__init__(AttrHolder(param_attr=param_attr))

        def __call__(self, input):
            return layers.row_conv(
                input=input,
                future_context_size=future_context_size,
                param_attr=self.attr_holder.param_attr,
                act=act)

    return RowConv_()


def layer_norm(**kwargs):
    raise NotImplementedError()


fluid_version = get_fluid_version()
if fluid_version >= 162 or fluid_version == 0:

    def batch_norm(act=None,
                   momentum=0.9,
                   epsilon=1e-05,
                   param_attr=None,
                   bias_attr=None,
                   data_layout='NCHW',
                   in_place=False,
                   name=None,
                   moving_mean_name=None,
                   moving_variance_name=None,
                   do_model_average_for_mean_and_var=False,
                   use_global_stats=False):
        """
        Return a function that creates a paddle.fluid.layers.batch_norm.

        """
        default_name = "batch_norm"
        param_attr = update_attr_name(name, default_name, param_attr, False)
        bias_attr = update_attr_name(name, default_name, bias_attr, True)
        moving_mean_attr = update_attr_name(
            name, default_name + "_moving_mean", None, False)
        moving_variance_attr = update_attr_name(
            name, default_name + "_moving_variance", None, False)

        class BatchNorm_(LayerFunc):
            def __init__(self):
                super(BatchNorm_, self).__init__(
                    AttrHolder(
                        param_attr=param_attr,
                        bias_attr=bias_attr,
                        moving_mean_attr=moving_mean_attr,
                        moving_variance_attr=moving_variance_attr))

            def __call__(self, input, is_test=False):
                return layers.batch_norm(
                    input=input,
                    act=act,
                    is_test=is_test,
                    momentum=momentum,
                    epsilon=epsilon,
                    param_attr=self.attr_holder.param_attr,
                    bias_attr=self.attr_holder.bias_attr,
                    data_layout=data_layout,
                    in_place=in_place,
                    name=name,
                    moving_mean_name=self.attr_holder.moving_mean_attr.name,
                    moving_variance_name=self.attr_holder.moving_variance_attr.
                    name,
                    do_model_average_for_mean_and_var=
                    do_model_average_for_mean_and_var,
                    use_global_stats=use_global_stats)

        return BatchNorm_()
else:

    def batch_norm(act=None,
                   momentum=0.9,
                   epsilon=1e-05,
                   param_attr=None,
                   bias_attr=None,
                   data_layout='NCHW',
                   in_place=False,
                   name=None,
                   moving_mean_name=None,
                   moving_variance_name=None,
                   do_model_average_for_mean_and_var=False,
                   fuse_with_relu=False,
                   use_global_stats=False):
        """
        Return a function that creates a paddle.fluid.layers.batch_norm.

        """
        default_name = "batch_norm"
        param_attr = update_attr_name(name, default_name, param_attr, False)
        bias_attr = update_attr_name(name, default_name, bias_attr, True)
        moving_mean_attr = update_attr_name(
            name, default_name + "_moving_mean", None, False)
        moving_variance_attr = update_attr_name(
            name, default_name + "_moving_variance", None, False)

        class BatchNorm_(LayerFunc):
            def __init__(self):
                super(BatchNorm_, self).__init__(
                    AttrHolder(
                        param_attr=param_attr,
                        bias_attr=bias_attr,
                        moving_mean_attr=moving_mean_attr,
                        moving_variance_attr=moving_variance_attr))

            def __call__(self, input, is_test=False):
                return layers.batch_norm(
                    input=input,
                    act=act,
                    is_test=is_test,
                    momentum=momentum,
                    epsilon=epsilon,
                    param_attr=self.attr_holder.param_attr,
                    bias_attr=self.attr_holder.bias_attr,
                    data_layout=data_layout,
                    in_place=in_place,
                    name=name,
                    moving_mean_name=self.attr_holder.moving_mean_attr.name,
                    moving_variance_name=self.attr_holder.moving_variance_attr.
                    name,
                    do_model_average_for_mean_and_var=
                    do_model_average_for_mean_and_var,
                    fuse_with_relu=fuse_with_relu,
                    use_global_stats=use_global_stats)

        return BatchNorm_()


def create_parameter(shape,
                     dtype,
                     name=None,
                     attr=None,
                     is_bias=False,
                     default_initializer=None):
    """
    Return a function that creates a paddle.fluid.layers.create_parameter.

    """
    param_attr = update_attr_name(name, "create_parameter", attr, False)

    class CreateParameter_(LayerFunc):
        def __init__(self):
            super(CreateParameter_, self).__init__(
                AttrHolder(param_attr=param_attr))

        def __call__(self):
            return layers.create_parameter(
                shape=shape,
                dtype=dtype,
                attr=self.attr_holder.param_attr,
                is_bias=is_bias,
                default_initializer=default_initializer)

    return CreateParameter_()



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six

__all__ = ['PiecewiseScheduler', 'LinearDecayScheduler']


class PiecewiseScheduler(object):
    """Set hyper parameters by a predefined step-based scheduler.
    """

    def __init__(self, scheduler_list):
        """Piecewise scheduler of hyper parameter.

        Args:
            scheduler_list: list of (step, value) pair. E.g. [(0, 0.001), (10000, 0.0005)]
        """
        assert len(scheduler_list) > 0

        for i in six.moves.range(len(scheduler_list) - 1):
            assert scheduler_list[i][0] < scheduler_list[i + 1][0], \
                    'step of scheduler_list should be incremental.'

        self.scheduler_list = scheduler_list

        self.cur_index = 0
        self.cur_step = 0
        self.cur_value = self.scheduler_list[0][1]

        self.scheduler_num = len(self.scheduler_list)

    def step(self, step_num=1):
        """Step step_num and fetch value according to following rule:

        Given scheduler_list: [(step_0, value_0), (step_1, value_1), ..., (step_N, value_N)],
        function will return value_K which satisfying self.cur_step >= step_K and self.cur_step < step_K+1

        Args:
            step_num (int): number of steps (default: 1)
        """
        assert isinstance(step_num, int) and step_num >= 1
        self.cur_step += step_num

        if self.cur_index < self.scheduler_num - 1:
            if self.cur_step >= self.scheduler_list[self.cur_index + 1][0]:
                self.cur_index += 1
                self.cur_value = self.scheduler_list[self.cur_index][1]

        return self.cur_value


class LinearDecayScheduler(object):
    """Set hyper parameters by a step-based scheduler with linear decay values.
    """

    def __init__(self, start_value, max_steps):
        """Linear decay scheduler of hyper parameter.
        Decay value linearly untill 0.
        
        Args:
            start_value (float): start value
            max_steps (int): maximum steps
        """
        assert max_steps > 0
        self.cur_step = 0
        self.max_steps = max_steps
        self.start_value = start_value

    def step(self, step_num=1):
        """Step step_num and fetch value according to following rule:

        return_value = start_value * (1.0 - (cur_steps / max_steps))

        Args:
            step_num (int): number of steps (default: 1)

        Returns:
            value (float): current value
        """
        assert isinstance(step_num, int) and step_num >= 1
        self.cur_step = min(self.cur_step + step_num, self.max_steps)

        value = self.start_value * (1.0 - (
            (self.cur_step * 1.0) / self.max_steps))

        return value



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
warnings.simplefilter('default')

import os
import paddle
import paddle.fluid as fluid
from parl.core.fluid import layers
from parl.core.agent_base import AgentBase
from parl.core.fluid.algorithm import Algorithm
from parl.utils import machine_info
from parl.utils import logger

__all__ = ['Agent']


class Agent(AgentBase):
    """
    | `alias`: ``parl.Agent``
    | `alias`: ``parl.core.fluid.agent.Agent``

    | Agent is one of the three basic classes of PARL.

    | It is responsible for interacting with the environment and collecting data for training the policy.
    | To implement a customized ``Agent``, users can:

      .. code-block:: python

        import parl

        class MyAgent(parl.Agent):
            def __init__(self, algorithm, act_dim):
                super(MyAgent, self).__init__(algorithm)
                self.act_dim = act_dim
      This class will initialize the neural network parameters automatically, and provides an executor for users to run the programs (self.fluid_executor).

    Attributes:
        fluid_executor (fluid.Executor): executor for running programs of the agent.
        alg (parl.algorithm): algorithm of this agent.

    Public Functions:
        - ``build_program`` (**abstract function**): build various programs for the agent to interact with outer environment.
        - ``get_weights``: return a Python dictionary containing all the parameters of self.alg.
        - ``set_weights``: copy parameters from ``set_weights()`` to this agent.
        - ``sample``: return a noisy action to perform exploration according to the policy.
        - ``predict``: return an action given current observation.
        - ``learn``: update the parameters of self.alg using the `learn_program` defined in `build_program()`.
        - ``save``: save parameters of the ``agent`` to a given path.
        - ``restore``: restore previous saved parameters from a given path.

    Todo:
        - allow users to get parameters of a specified model by specifying the model's name in ``get_weights()``.

    """

    def __init__(self, algorithm):
        """Build programs by calling the method ``self.build_program()`` and run initialization function of ``fluid.default_startup_program()``.

        Args:
            algorithm (parl.Algorithm): an instance of `parl.Algorithm`. This algorithm is then passed to `self.alg`.
        """

        assert isinstance(algorithm, Algorithm)
        super(Agent, self).__init__(algorithm)

        self.gpu_id = 0 if machine_info.is_gpu_available() else -1

        self.build_program()

        if machine_info.is_xpu_available():
            self.place = fluid.XPUPlace(int(os.getenv("FLAGS_selected_xpus")))
        elif machine_info.is_gpu_available():
            self.place = fluid.CUDAPlace(0)
        else:
            self.place = fluid.CPUPlace()
        self.fluid_executor = fluid.Executor(self.place)
        self.fluid_executor.run(fluid.default_startup_program())

    def build_program(self):
        """Build various programs here with the
        learn, predict, sample functions of the algorithm.

        Note:
            | Users **must** implement this function in an ``Agent``.
            | This function will be called automatically in the initialization function.

        To build a program, you must do the following:
            a. Create a fluid program with ``fluid.program_guard()``;
            b. Define data layers for feeding the data;
            c. Build various programs(e.g., learn_program, predict_program) with data layers defined in step b.

        Example:

        .. code-block:: python

            self.pred_program = fluid.Program()

            with fluid.program_guard(self.pred_program):
                obs = layers.data(
                    name='obs', shape=[self.obs_dim], dtype='float32')
                self.act_prob = self.alg.predict(obs)


        """
        raise NotImplementedError

    def get_model_ids(self):
        """Get all model ids of the self.alg in the agent.

        Returns:
            List of model_id 
        """
        return self.alg.get_model_ids()

    @property
    def model_ids(self):
        return self.get_model_ids()

    def learn(self, *args, **kwargs):
        """The training interface for ``Agent``.
        This function feeds the training data into the learn_program defined in ``build_program()``.
        """
        raise NotImplementedError

    def predict(self, *args, **kwargs):
        """Predict an action when given the observation of the environment.

        This function feeds the observation into the prediction program defined in ``build_program()``. It is often used in the evaluation stage.
        """
        raise NotImplementedError

    def sample(self, *args, **kwargs):
        """Return an action with noise when given the observation of the environment.

        In general, this function is used in train process as noise is added to the action to preform exploration.

        """
        raise NotImplementedError

    def save(self, save_path=None, program=None):
        """Save parameters.

        Args:
            save_path(str): a directory where to save the parameters.
            program(fluid.Program): program that describes the neural network structure. If None, will all program.

        Raises:
            Error: if program does not exist

        Example:

        .. code-block:: python

            agent = AtariAgent()
            agent.save('./model_dir')
            agent.save('./model_dir', program=agent.learn_program)

        """
        assert save_path is not None, 'please specify `save_path` '
        if os.path.isfile(save_path):
            raise Exception('can not save to {}, it is a file, not directory'.
                            format(save_path))
        if not os.path.exists(save_path):
            os.makedirs(save_path)
        all_programs = [(key, val) for (key, val) in self.__dict__.items()
                        if (isinstance(val, fluid.framework.Program)
                            or isinstance(val, fluid.compiler.CompiledProgram))
                        ]

        if program:
            filename = None
            for (name, prog) in all_programs:
                if program == prog:
                    filename = name
                    break
            if filename is None:
                raise Exception('can not find program {}.'.format(program))
            fluid.io.save_params(
                executor=self.fluid_executor,
                dirname=save_path,
                main_program=program,
                filename=filename)
        else:
            for (filename, program) in all_programs:
                if isinstance(program, fluid.framework.Program) or \
                        isinstance(program, fluid.compiler.CompiledProgram):
                    fluid.io.save_params(
                        executor=self.fluid_executor,
                        dirname=save_path,
                        main_program=program,
                        filename=filename)

    def restore(self, save_path=None, program=None):
        """Restore previously saved parameters from save_path. 

        Args:
            save_path(str): path where parameters were previously saved.
            program(fluid.Program): program that describes the neural network structure. If None, will restore all program.

        Raises:
            Error: if save_path does not exist or can not find the specific program file in save_path.

        Example:

        .. code-block:: python

            agent = AtariAgent()
            agent.save('./model_dir')
            agent.restore('./model_dir')

        """
        assert save_path is not None, 'please specify `save_path` '
        if not os.path.exists(save_path):
            raise Exception(
                'can not restore from {}, directory does not exists'.format(
                    save_path))
        if os.path.isfile(save_path):
            raise Exception(
                'can not restore from {}, it is a file, not directory'.format(
                    save_path))
        all_programs = [(key, val) for (key, val) in self.__dict__.items()
                        if (isinstance(val, fluid.framework.Program)
                            or isinstance(val, fluid.compiler.CompiledProgram))
                        ]

        if program:
            filename = None
            for (name, prog) in all_programs:
                if program == prog:
                    filename = name
                    break
            if filename is None:
                raise Exception('can not find the program to restore.')
            if not os.path.isfile('{}/{}'.format(save_path, filename)):
                raise Exception('{}/{} does not exits'.format(
                    save_path, filename))
            if type(program) is fluid.compiler.CompiledProgram:
                program = program._init_program
            fluid.io.load_params(
                executor=self.fluid_executor,
                dirname=save_path,
                main_program=program,
                filename=filename)
        else:
            programs_list = [kv[0] for kv in all_programs]
            exist_files = os.listdir(save_path)
            if len(programs_list) != len(exist_files):
                raise Exception(
                    'expected to restore {} model file under directory {}: {}, but {} files are found: {}.'
                    .format(
                        len(programs_list), save_path, programs_list,
                        len(exist_files), exist_files))
            for (filename, program) in all_programs:
                if not os.path.isfile('{}/{}'.format(save_path, filename)):
                    raise Exception('{}/{} does not exits'.format(
                        save_path, filename))
                if type(program) is fluid.compiler.CompiledProgram:
                    program = program._init_program

                fluid.io.load_params(
                    executor=self.fluid_executor,
                    dirname=save_path,
                    main_program=program,
                    filename=filename)



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

XPARL_RESERVED_PREFIX = "_xparl"
RESERVED_NAME_ERROR_STR = "Name starting with `_xparl` is the reserved name in xparl, please don't use the name `{}`."


def proxy_wrapper_func(remote_wrapper):
    '''
    The 'proxy_wrapper_func' is defined on the top of class 'RemoteWrapper'
    in order to set and get attributes of 'remoted_wrapper' and the corresponding 
    remote models individually. 

    This decorator function allows the user access the attributes of remote objects 
    like local objects.

    For example:
        ```python
        import parl

        @parl.remote_class()
        class Actor(object):
            def __init__(self):
                self.arg1 = 0

            def func(self):
                # do something
                return 0

        parl.connect("localhost:8010")

        actor = Actor()
        print(actor.arg1)
        actor.arg2 = 10
        print(actor.func())
        ```
    '''

    original_class = remote_wrapper._original
    max_memory = remote_wrapper._max_memory

    class ProxyWrapper(object):
        def __init__(self, *args, **kwargs):
            for key in kwargs:
                assert not key.startswith(
                    XPARL_RESERVED_PREFIX), RESERVED_NAME_ERROR_STR.format(key)

            # The following variables will be used in the RemoteWrapper, so we put them
            # into the kwargs.
            kwargs['_xparl_remote_class'] = original_class
            kwargs['_xparl_remote_class_max_memory'] = max_memory

            self._xparl_remote_wrapper_obj = remote_wrapper(*args, **kwargs)
            for key in self._xparl_remote_wrapper_obj.get_attrs():
                assert not key.startswith(
                    XPARL_RESERVED_PREFIX), RESERVED_NAME_ERROR_STR.format(key)

        def __getattr__(self, attr):
            return self._xparl_remote_wrapper_obj.get_remote_attr(attr)

        def __setattr__(self, attr, value):
            if attr.startswith(XPARL_RESERVED_PREFIX):
                super(ProxyWrapper, self).__setattr__(attr, value)
            else:
                self._xparl_remote_wrapper_obj.set_remote_attr(attr, value)

    return ProxyWrapper



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import paddle.fluid as fluid
import os


def compile(program, loss=None):
    """ transfer the program into a new program that runs in multi-cpus or multi-gpus.
    This function uses the `fluid.compiler.CompiledProgram` to transfer the program.
    For more detail about speeding the program, please visit 
    "https://www.paddlepaddle.org.cn/documentation/docs/zh/1.5/user_guides/howto/training/single_node.html#id7"

    Args:
        program(fluid.Program): a normal fluid program.
        loss_name(str): Optional. The loss tensor of a trainable program. Set it to None if you are transferring a prediction or evaluation program.
    """
    loss_name = None
    if loss is not None:
        assert isinstance(
            loss, fluid.framework.
            Variable), 'type of loss is expected to be a fluid tensor'
        loss_name = loss.name
    # TODO: after solving the learning rate issue that occurs in training A2C algorithm, set it to 3.
    os.environ['CPU_NUM'] = '1'
    exec_strategy = fluid.ExecutionStrategy()
    exec_strategy.num_threads = 3 * 4

    build_strategy = fluid.BuildStrategy()
    build_strategy.reduce_strategy = fluid.BuildStrategy.ReduceStrategy.Reduce

    compiled_program = fluid.compiler.CompiledProgram(
        program).with_data_parallel(
            loss_name=loss_name,
            exec_strategy=exec_strategy,
            build_strategy=build_strategy)
    compiled_program._init_program = program
    return compiled_program



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import errno
import logging
import os
import os.path
import sys
from termcolor import colored
import shutil
from datetime import datetime

__all__ = [
    'set_dir', 'get_dir', 'set_level', 'auto_set_dir', 'add_stdout_handler',
    'remove_handler'
]

# globals: logger file and directory:
LOG_DIR = None
_FILE_HANDLER = None


def _makedirs(dirname):
    assert dirname is not None
    if dirname == '' or os.path.isdir(dirname):
        return
    try:
        os.makedirs(dirname)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise e


def _get_time_str():
    return datetime.now().strftime('%m%d-%H%M%S')


class _Formatter(logging.Formatter):
    def format(self, record):
        msg = '%(message)s'
        if record.levelno == logging.WARNING:
            date = colored(
                '[%(asctime)s %(threadName)s @%(filename)s:%(lineno)d]',
                'yellow')
            fmt = date + ' ' + colored(
                'WRN', 'yellow', attrs=['blink']) + ' ' + msg
        elif record.levelno == logging.ERROR or record.levelno == logging.CRITICAL:
            date = colored(
                '[%(asctime)s %(threadName)s @%(filename)s:%(lineno)d]', 'red')
            fmt = date + ' ' + colored(
                'WRN', 'yellow', attrs=['blink']) + ' ' + msg
            fmt = date + ' ' + colored(
                'ERR', 'red', attrs=['blink', 'underline']) + ' ' + msg
        elif record.levelno == logging.DEBUG:
            date = colored(
                '[%(asctime)s %(threadName)s @%(filename)s:%(lineno)d]',
                'blue')
            fmt = date + ' ' + colored(
                'DEBUG', 'blue', attrs=['blink']) + ' ' + msg
        else:
            date = colored(
                '[%(asctime)s %(threadName)s @%(filename)s:%(lineno)d]',
                'green')
            fmt = date + ' ' + msg
        if hasattr(self, '_style'):
            # Python3 compatibility
            self._style._fmt = fmt
        self._fmt = fmt
        return super(_Formatter, self).format(record)


def _getlogger():
    logger = logging.getLogger('PARL')
    logger.propagate = False
    logger.setLevel(logging.DEBUG)

    if 'DEBUG' in os.environ:
        handler = logging.FileHandler('parl_debug.log')
        handler.setFormatter(_Formatter(datefmt='%m-%d %H:%M:%S'))
        logger.addHandler(handler)
        return logger

    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(_Formatter(datefmt='%m-%d %H:%M:%S'))
    logger.addHandler(handler)
    return logger


_logger = _getlogger()
_LOGGING_METHOD = [
    'info', 'warning', 'error', 'critical', 'warn', 'exception', 'debug',
    'setLevel'
]

# export logger functions
for func in _LOGGING_METHOD:
    locals()[func] = getattr(_logger, func)
    __all__.append(func)

# export Level information
_LOGGING_LEVEL = ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']
for level in _LOGGING_LEVEL:
    locals()[level] = getattr(logging, level)
    __all__.append(level)


def _set_file(path):
    global _FILE_HANDLER, _logger
    if os.path.isfile(path):
        try:
            os.remove(path)
        except OSError:
            pass
    hdl = logging.FileHandler(filename=path, encoding='utf-8', mode='w')
    hdl.setFormatter(_Formatter(datefmt='%m-%d %H:%M:%S'))

    _FILE_HANDLER = hdl
    _logger.addHandler(hdl)


def set_level(level):
    global _logger, LOG_DIR
    # To set level, need create new handler
    if LOG_DIR is not None:
        set_dir(get_dir())
    _logger.setLevel(level)


def set_dir(dirname):
    global LOG_DIR, _FILE_HANDLER, _logger
    if _FILE_HANDLER:
        # unload and close the old file handler, so that we may safely delete the logger directory
        _logger.removeHandler(_FILE_HANDLER)
        _FILE_HANDLER.close()
        del _FILE_HANDLER

    shutil.rmtree(dirname, ignore_errors=True)
    _makedirs(dirname)
    LOG_DIR = dirname
    _set_file(os.path.join(dirname, 'log.log'))


def auto_set_dir(action=None):
    """Set the global logging directory automatically. The default path is "./train_log/{scriptname}". "scriptname" is the name of the main python file currently running"

    Note: This function references `https://github.com/tensorpack/tensorpack/blob/master/tensorpack/utils/logger.py#L93`

    Args:
        dir_name(str): log directory
        action(str): an action of ["k","d","q"] to be performed
            when the directory exists. Will ask user by default.
                "d": delete the directory. Note that the deletion may fail when
                the directory is used by tensorboard.
                "k": keep the directory. This is useful when you resume from a
                previous training and want the directory to look as if the
                training was not interrupted.
                Note that this option does not load old models or any other
                old states for you. It simply does nothing.

    Returns:
        dirname(str): log directory used in the global logging directory.
    """
    mod = sys.modules['__main__']
    if hasattr(mod, '__file__'):
        basename = os.path.basename(mod.__file__)
    else:
        basename = ''
    dirname = os.path.join('train_log', basename[:basename.rfind('.')])
    dirname = os.path.normpath(dirname)

    global LOG_DIR, _FILE_HANDLER
    if _FILE_HANDLER:
        # unload and close the old file handler, so that we may safely delete the logger directory
        _logger.removeHandler(_FILE_HANDLER)
        del _FILE_HANDLER

    def dir_nonempty(dirname):
        # If directory exists and nonempty (ignore hidden files), prompt for action
        return os.path.isdir(dirname) and len(
            [x for x in os.listdir(dirname) if x[0] != '.'])

    if dir_nonempty(dirname):
        if not action:
            _logger.warning("""\
Log directory {} exists! Use 'd' to delete it. """.format(dirname))
            _logger.warning("""\
If you're resuming from a previous run, you can choose to keep it.
Press any other key to exit. """)
        while not action:
            action = input("Select Action: k (keep) / d (delete) / q (quit):"
                           ).lower().strip()
        act = action
        if act == 'd':
            shutil.rmtree(dirname, ignore_errors=True)
            if dir_nonempty(dirname):
                shutil.rmtree(dirname, ignore_errors=False)
        elif act == 'n':
            dirname = dirname + _get_time_str()
            info("Use a new log directory {}".format(dirname))  # noqa: F821
        elif act == 'k':
            pass
        else:
            raise OSError("Directory {} exits!".format(dirname))
    LOG_DIR = dirname
    _makedirs(dirname)
    _set_file(os.path.join(dirname, 'log.log'))
    return dirname


def get_dir():
    return LOG_DIR


def add_stdout_handler():
    global _logger
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(_Formatter(datefmt='%m-%d %H:%M:%S'))
    _logger.addHandler(handler)
    return handler


def remove_handler(handler):
    global _logger
    _logger.removeHandler(handler)


# Will save log to log_dir/main_file_name/log.log by default

mod = sys.modules['__main__']
if hasattr(mod, '__file__') and 'XPARL' not in os.environ:
    _logger.info("Argv: " + ' '.join(sys.argv))
